// Include files
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "TComMath.h"
#include "TCom1Dlut.h"

Bool  TCom1Dlut::matSolve( Double **mat, Double *Y, Int nbp, Double *X )
{
//eliminationGaussJordan
/*
  int     n=4, o=0; 
  double  e[5][4]; 
  Double  GainOffsetMatrix[3][4];
*/
  int     n=nbp, o=0;
  Double  **e = TComMath::allocateArray2D( nbp, nbp+1 );
  Bool    bVerbose  = true;
  Bool    bOk       = true;

  Double  max = 0.0;
  for (int i=0;i<n;i++){   
    for (int p=0;p<n;p++){ 
      e[p][i] = mat[i][p] ; 
      o = p;

      max = (fabs(e[p][i]) > max) ? fabs(e[p][i]) : max ;
    } 
    e[n][i] = Y[i] ; 
    max = (fabs(Y[i]) > max) ? fabs(Y[i]) : max ;
  } 
  //TComMath::displayArray2D( "e[][]", e, nbp, nbp+1 );
  /*********method*************/

  TComMath::displayArray2D( "e", e, nbp, nbp+1 );
  printf("max = %g\n", max );
  TComMath::matMult( e, 0.001*max, nbp, nbp+1 );
  TComMath::displayArray2D( "e normalized", e, nbp, nbp+1 );

  // on a saisi les facteurs des equations ds e[][] 
  int y=0; 
  double var1=0,var2=0; 
  double temp; 
  int a,t; 
  for(int x=0;x<n-1;x++){
    for(a=1+x;a<n;a++){ 
      temp=e[x][a]; 
      for (t=x;t<n+1;t++){ 
        e[t][a]=e[t][a]*e[x][x]-e[t][x]*temp; 
        /// triangulation du systeme. 
      } 
    } 
  } 

  TComMath::displayArray2D( "e", e, nbp, nbp+1 );

  if ( e[n-1][n-1] == 0.0 ) {
    if ( bVerbose ) printf("matSolve():\tAie.\n");
    bOk = false;
    return bOk ;
  }

  //Passons au remplacage..  
  X[n-1]=e[n][n-1]/e[n-1][n-1]; 
  e[n][n-1]=0; 
  e[n-1][n-1]=0; 
  for (int ligne=1;ligne<=n;ligne++){ 
    for (int sol=2;sol<=n;sol++){ 
      e[n-ligne][n-sol]*=X[n-ligne]; 
      e[n][n-sol]-=e[n-ligne][n-sol]; 
      e[n-ligne][n-sol]=0; 
    } 
  if ( n-(ligne+1) >= 0 ) // bug fixing PB
    X[n-(ligne+1)]=e[n][n-(ligne+1)]/e[n-(ligne+1)][n-(ligne+1)]; 
  }

  //TComMath::displayArray2D( "e[][]", e, nbp, nbp+1 );

  TComMath::deleteArray2D( e, nbp+1 );

  return true ;
}

TCom1Dlut::TCom1Dlut( Int iNbpCode, Int iInputBitDepth, Int iOutputBitDepth, Bool bVerbose )
{
  m_bVerbose        = bVerbose;
  m_iNbpCode        = iNbpCode;
  m_iInputBitDepth  = iInputBitDepth;
  m_iOutputBitDepth = iOutputBitDepth;

  m_nbp = 1 + ( 1 << ( m_iNbpCode - 1 ) );

  m_lut = (Double *)malloc( sizeof(Double) * m_nbp );
  init();

  m_Y     = (Double *)malloc( sizeof(Double) * m_nbp );
  m_X     = (Double *)malloc( sizeof(Double) * m_nbp );
  m_mat   = TComMath::allocateArray2D( m_nbp, m_nbp );
  m_nbPts = (Int *)malloc( sizeof(Int) * m_nbp );
  razData();
}

TCom1Dlut::~TCom1Dlut()
{
  free(m_lut);

  free(m_Y);
  free(m_X);
  TComMath::deleteArray2D( m_mat, m_nbp );
}

Void  TCom1Dlut::razData()
{
  for ( Int y=0 ; y<m_nbp ; y++ ) {
    m_Y[y]      = 0.0;
    m_X[y]      = 0.0;
    m_nbPts[y]  = 0;
    for ( Int x=0 ; x<m_nbp ; x++ ) {
      m_mat[y][x] = 0.0;
    }
  }
}

Void	TCom1Dlut::init()
{
  for ( Int i=0 ; i<m_nbp ; i++ )
    m_lut[i] = (Double)( i << (m_iOutputBitDepth - m_iNbpCode + 1) );
}

Void	TCom1Dlut::appendColorSample( Double dY0, Double dY1 )
{
  dY0 = TComMath::xClipD( dY0, (Double)( (1<<m_iInputBitDepth)  ) );
  dY1 = TComMath::xClipD( dY1, (Double)( (1<<m_iOutputBitDepth) ) );
  Int sY0     = (Int)dY0;
  Int iShift  = m_iInputBitDepth - m_iNbpCode + 1;
	Int	idx     = sY0 >> iShift;

  Double  s   = ( dY0 - (Double)(idx<<iShift) ) / (Double)(1 << iShift);
  Double  s2  = s * s;

  //dY0 = dY0 / (Double)(1 << iShift);
  //Double  dY02= dY0 * dY0;
  dY1 = dY1 / (Double)(1 << (m_iOutputBitDepth - m_iNbpCode + 1));

  m_Y[idx]    += dY1 * (1.0 - s) ;
  if ( (idx+1) < m_nbp )
    m_Y[idx+1]  += dY1 * s ;

  m_mat[idx][idx]     +=  s2 - 2.0*s + 1.0 ;
  if ( (idx+1) < m_nbp ) {
    m_mat[idx][idx+1]   +=  -s2 + s ;
    m_mat[idx+1][idx]   +=  -s2 + s ;
    m_mat[idx+1][idx+1] +=  s2 ;
  }
  m_nbPts[idx]++;
}

Void	TCom1Dlut::computeLut()
{
  Double  sum = 0.0;
  for ( Int n=0 ; n<m_nbp ; n++ ) {
    if ( m_nbPts[n]==0 ) {
      for ( Int t=0 ; t<1 ; t++ )
        appendColorSample( (Double)(n<<(m_iInputBitDepth - m_iNbpCode + 1)), (Double)(n<<(m_iOutputBitDepth - m_iNbpCode + 1)) );
    }
    sum += (Double)m_nbPts[n];
  }

  //TComMath::displayArray2D( "", m_mat, m_nbp, m_nbp );

  Double  *tmp_lut  = (Double *)malloc( sizeof(Double) * m_nbp );
  //Bool bOk = matSolve( m_mat, m_Y, m_nbp, tmp_lut );  // worse and less stable than "TComMath::matSolve()"
  Bool bOk = TComMath::matSolve( m_mat, m_Y, m_nbp, tmp_lut );

  if ( bOk )
  for ( Int n=0 ; n<m_nbp ; n++ )
    m_lut[n] = tmp_lut[n] * (Double)(1 << (m_iOutputBitDepth - m_iNbpCode + 1));

  delete tmp_lut ;
}

Void	TCom1Dlut::displayLut( Char *lutName, Int iComp, Bool bVerbose )
{
  if ( bVerbose ) printf("%s[%d] = ", lutName, iComp);
  for ( Int n=0 ; n<m_nbp ; n++ ) {
    //if ( bVerbose ) printf("%d(%d)\t", (Int)( 0.5 + m_lut1D_A[iComp][n].a*(Double)(n << (8 - m_shift_div)) + m_lut1D_A[iComp][n].b ), m_lut1D_A[iComp][n].nbPts );
    if ( bVerbose ) printf("%g\t", m_lut[n] );
  }
  if ( bVerbose ) printf("\n");
}
