/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2014, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef __TCOM1DLUT__
#define __TCOM1DLUT__

// Include files
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "TypeDef.h"
#include "TComMath.h"


class TCom1Dlut
{
  Bool    m_bVerbose;
  Int 		m_iNbpCode;
  Int 		m_nbp;
  Int     m_iInputBitDepth;
  Int     m_iOutputBitDepth;
	Double	*m_lut;
	
  //[tabY] = [matA][Lx]
  Double  *m_Y, *m_X, **m_mat;
  Int     *m_nbPts;

  Bool  matSolve( Double **mat, Double *Y, Int nbp, Double *X );

public:
  TCom1Dlut::TCom1Dlut( Int iNbpCode, Int iInputBitDepth, Int iOutputBitDepth, Bool bVerbose=true );
  TCom1Dlut::~TCom1Dlut();

  Void  init();
  Void  razData();
  Void  appendColorSample( Double sY0, Double sY1 );
  Void  computeLut();
  Void  displayLut( Char *lutName, Int iComp, Bool bVerbose );

  inline	Double	modelLut( Pel sY0, Int inputBitDepth, Bool bClipOut=false );
  inline	Double	modelLut( Double dY0, Int inputBitDepth, Bool bClipOut=false );

  inline  Double  getLut( Int n ) { return m_lut[n];  };
  inline  Int     getNbPts()      { return m_nbp;     };
};

inline	Double	TCom1Dlut::modelLut( Pel sY0, Int inputBitDepth, Bool bClipOut )
{ 
  Int iShift  = inputBitDepth - m_iNbpCode + 1;
	Int	idx	= sY0 >> iShift;

  Double  val = m_lut[idx+1] * ( (Double)sY0 - (Double)(idx<<iShift) ) + m_lut[idx] * ( (Double)((idx+1)<<iShift) - (Double)sY0 );

  val /= (Double)( 1 << iShift );

  if ( bClipOut ) {
    TComMath::xClipD( val, (Double)(1<<m_iOutputBitDepth) - 1.0 );
  }

  return val ;
}

inline	Double	TCom1Dlut::modelLut( Double dY0, Int inputBitDepth, Bool bClipOut )
{ 
  Int iShift  = inputBitDepth - m_iNbpCode + 1;
	Int	idx	= (Int)dY0 >> iShift;

  Double  val = m_lut[idx+1] * ( dY0 - (Double)(idx<<iShift) ) + m_lut[idx] * ( (Double)((idx+1)<<iShift) - dY0 );

  val /= (Double)( 1 << iShift );

  if ( bClipOut ) {
    TComMath::xClipD( val, (Double)(1<<m_iOutputBitDepth) - 1.0 );
  }

  return val ;
}

#endif