// Include files
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "TComColorMappingAlgo.h"
#include "TCom1DlutGO.h"
#include "TComMath.h"


static  Void  assert( Char *comment, Bool bTest )
{
  if ( !bTest ) {
    printf("assertion failed:\t%s\n", comment);
    exit(0);
  }
}

/*-----------------------------------------
|E| | A B | |a|
| |=|     | | |
|F| | C D | |b|

|a| | m00 m01 | |E|
| |=|         | | |
|b| | m10 m11 | |F|

-----------------------------------------*/
static Void computeLinearModel( linearModel* m )
{
  Double  m00, m01, m10, m11;
  Double  det = (m->A*m->D - m->B*m->C) ;

  if ( det == 0.0 ) return;

  m00 =   m->D / det;
  m01 = - m->B / det;
  m10 = - m->C / det;
  m11 =   m->A / det;

  Double  new_a = m00*m->E + m01*m->F;
  Double  new_b = m10*m->E + m11*m->F;

  //Double  alpha = 0.5;
  Double  alpha = 1.0;
  m->a  = (1.0 - alpha) * m->a  + alpha * new_a;
  m->b  = (1.0 - alpha) * m->b  + alpha * new_b;
}

//static Void compute_GO_model( ColorData* pColorData, Double outGainOffsetMatrix[3][4], Bool m_bVerbose )
static Void compute_GO_model( ColorData* pColorData, Double **outGainOffsetMatrix, Bool m_bVerbose )
{
//eliminationGaussJordan
  int     n=4, o=0; 
  double  e[5][4]; 
  Double  GainOffsetMatrix[3][4];

  pColorData->bOk = true;

  for (int row=0; row<=2; row++)
  {
    for (int i=0;i<n;i++){   
      for (int p=0;p<n;p++){ 
        e[p][i] = pColorData->dtchCoeffMatrix [i][p] ; 
        o = p;
      } 
        e[n][i] = pColorData->dtchCoeffMatrix [i][o+row+1] ; 
    } 
    /*********method*************/
 
    // on a saisi les facteurs des equations ds e[][] 
    int y=0; 
    double var1=0,var2=0; 
    double temp; 
    int a,t; 
    for(int x=0;x<n-1;x++){
      for(a=1+x;a<n;a++){ 
        temp=e[x][a]; 
        for (t=x;t<n+1;t++){ 
          e[t][a]=e[t][a]*e[x][x]-e[t][x]*temp; 
          /// triangulation du systeme. 
        } 
      } 
    } 

    if ( e[n-1][n-1] == 0.0 ) {
      if ( m_bVerbose ) printf("Aie.\n");
      pColorData->bOk = false;
      return;
    }

    //Passons au remplacage..  
    GainOffsetMatrix[row][n-1]=e[n][n-1]/e[n-1][n-1]; 
    e[n][n-1]=0; 
    e[n-1][n-1]=0; 
    for (int ligne=1;ligne<=n;ligne++){ 
      for (int sol=2;sol<=n;sol++){ 
        e[n-ligne][n-sol]*=GainOffsetMatrix[row][n-ligne]; 
        e[n][n-sol]-=e[n-ligne][n-sol]; 
        e[n-ligne][n-sol]=0; 
      } 
    if ( n-(ligne+1) >= 0 ) // bug fixing PB
      GainOffsetMatrix[row][n-(ligne+1)]=e[n][n-(ligne+1)]/e[n-(ligne+1)][n-(ligne+1)]; 
    }
  }
  
  //Double  alpha = 0.5;
  Double  alpha = 1.0;
  for ( int i=0 ; i<3 ; i++ )
  for ( int j=0 ; j<4 ; j++ )
    outGainOffsetMatrix[i][j]  = (1.0 - alpha) * outGainOffsetMatrix[i][j] + alpha * GainOffsetMatrix[i][j];

  if ( m_bVerbose ) {
    for ( int i=0 ; i<3 ; i++ )
      printf("\t%g\t%g\t%g\t\t%g\n", outGainOffsetMatrix[i][0], outGainOffsetMatrix[i][1], outGainOffsetMatrix[i][2], outGainOffsetMatrix[i][3] );
  }
}

static  Void  appendColorData_GO( ColorData* pColorData, Pel sY0, Pel sU0, Pel sV0, Pel sY1, Pel sU1, Pel sV1 )
{
  Double  dy_709    = (Double)sY0;
  Double  dcb_709   = (Double)sU0;
  Double  dcr_709   = (Double)sV0;
  Double  dy_2020   = (Double)sY1;
  Double  dcb_2020  = (Double)sU1;
  Double  dcr_2020  = (Double)sV1;

  pColorData->dtchCoeffMatrix[0][0] += dy_709*dy_709;
  pColorData->dtchCoeffMatrix[0][1] += dy_709*dcb_709;
  pColorData->dtchCoeffMatrix[0][2] += dy_709*dcr_709;
  pColorData->dtchCoeffMatrix[0][3] += dy_709;
  pColorData->dtchCoeffMatrix[0][4] += dy_2020*dy_709;
  pColorData->dtchCoeffMatrix[0][5] += dcb_2020*dy_709;
  pColorData->dtchCoeffMatrix[0][6] += dcr_2020*dy_709;

  pColorData->dtchCoeffMatrix[1][0] += dy_709*dcb_709;
  pColorData->dtchCoeffMatrix[1][1] += dcb_709*dcb_709;
  pColorData->dtchCoeffMatrix[1][2] += dcb_709*dcr_709;
  pColorData->dtchCoeffMatrix[1][3] += dcb_709;
  pColorData->dtchCoeffMatrix[1][4] += dy_2020*dcb_709;
  pColorData->dtchCoeffMatrix[1][5] += dcb_2020*dcb_709;
  pColorData->dtchCoeffMatrix[1][6] += dcr_2020*dcb_709;

  pColorData->dtchCoeffMatrix[2][0] += dy_709*dcr_709;
  pColorData->dtchCoeffMatrix[2][1] += dcb_709*dcr_709;
  pColorData->dtchCoeffMatrix[2][2] += dcr_709*dcr_709;
  pColorData->dtchCoeffMatrix[2][3] += dcr_709;
  pColorData->dtchCoeffMatrix[2][4] += dy_2020*dcr_709;
  pColorData->dtchCoeffMatrix[2][5] += dcb_2020*dcr_709;
  pColorData->dtchCoeffMatrix[2][6] += dcr_2020*dcr_709;
  
  pColorData->dtchCoeffMatrix[3][0] += dy_709;
  pColorData->dtchCoeffMatrix[3][1] += dcb_709;
  pColorData->dtchCoeffMatrix[3][2] += dcr_709;
  //pColorData->dtchCoeffMatrix[3][3] = ipic_size;
  (pColorData->dtchCoeffMatrix[3][3])++;
  pColorData->dtchCoeffMatrix[3][4] += dy_2020;
  pColorData->dtchCoeffMatrix[3][5] += dcb_2020;
  pColorData->dtchCoeffMatrix[3][6] += dcr_2020;

  (pColorData->nbPts)++;
}

static  Void  appendColorData_1DLut( ColorData* pColorData, linearModel* pLinearModel, Int iComp, Pel sY0, Pel sU0, Pel sV0, Pel sY1, Pel sU1, Pel sV1 )
{
  Double  m00 = pColorData->GainOffsetMatrix[0][0];
  Double  m01 = pColorData->GainOffsetMatrix[0][1];
  Double  m02 = pColorData->GainOffsetMatrix[0][2];
  Double  m03 = pColorData->GainOffsetMatrix[0][3];

  Double  m10 = pColorData->GainOffsetMatrix[1][0];
  Double  m11 = pColorData->GainOffsetMatrix[1][1];
  Double  m12 = pColorData->GainOffsetMatrix[1][2];
  Double  m13 = pColorData->GainOffsetMatrix[1][3];

  Double  m20 = pColorData->GainOffsetMatrix[2][0];
  Double  m21 = pColorData->GainOffsetMatrix[2][1];
  Double  m22 = pColorData->GainOffsetMatrix[2][2];
  Double  m23 = pColorData->GainOffsetMatrix[2][3];

  Double  dY0 = (Double)sY0;
  Double  dU0 = (Double)sU0;
  Double  dV0 = (Double)sV0;
  Double  dY1 = (Double)sY1;
  Double  dU1 = (Double)sU1;
  Double  dV1 = (Double)sV1;

  //Double  wY = 0.75; // if > 1.0, it will favour Y in the quadratic error...
  Double  wY = 1.0; // if > 1.0, it will favour Y in the quadratic error...
  Bool  bSeparateLutOptimization = false;

  if ( iComp == 0 ) { // Luma
    // sU0 supposed to L( sU0 ) 
    // sV0 supposed to L( sV0 ) 
    if ( bSeparateLutOptimization ) {
      m10 = 0.0;
      m20 = 0.0;
    }

    Double  m00_pow2  = m00 * m00 ;
    Double  m10_pow2  = m10 * m10 ;
    Double  m20_pow2  = m20 * m20 ;

    Double  temp0 = 2.0 * dY0 * ( wY*m00_pow2 + m10_pow2 + m20_pow2 );
    Double  temp1 = 2.0 *       ( wY*m00_pow2 + m10_pow2 + m20_pow2 );
    Double  temp2 = 2.0 *       ( wY*m00 * ( dY1 - m01 * dU0 - m02 * dV0 - m03 ) + m10 * ( dU1 - m11 * dU0 - m12 * dV0 - m13 ) + m20 * ( dV1 - m21 * dU0 - m22 * dV0 - m23 ) );
/*
    pLinearModel->A += 2.0 * (Double)( sY0*sY0 ) * ( m00_pow2 + m10_pow2 + m20_pow2 );
    pLinearModel->B += 2.0 * (Double)(     sY0 ) * ( m00_pow2 + m10_pow2 + m20_pow2 );
    pLinearModel->E += 2.0 * (Double)(     sY0 ) * ( m00 * ( sY1 - m01 * sU0 - m02 * sV0 - m03 ) + m10 * ( sU1 - m11 * sU0 - m12 * sV0 - m13 ) + m20 * ( sV1 - m21 * sU0 - m22 * sV0 - m23 ) );

    pLinearModel->C += 2.0 * (Double)( sY0     ) * ( m00_pow2 + m10_pow2 + m20_pow2 );
    pLinearModel->D += 2.0 *                       ( m00_pow2 + m10_pow2 + m20_pow2 );
    pLinearModel->F += 2.0 *                       ( m00 * ( sY1 - m01 * sU0 - m02 * sV0 - m03 ) + m10 * ( sU1 - m11 * sU0 - m12 * sV0 - m13 ) + m20 * ( sV1 - m21 * sU0 - m22 * sV0 - m23 ) );
*/
    pLinearModel->A += dY0 * temp0;
    pLinearModel->B += dY0 * temp1;
    pLinearModel->E += dY0 * temp2;

    pLinearModel->C += temp0;
    pLinearModel->D += temp1;
    pLinearModel->F += temp2;
  }
  else if ( iComp == 1 ) { // Cb
    if ( bSeparateLutOptimization ) {
      m01 = 0.0;
      m21 = 0.0;
    }

    dY0 *= wY;
    dY1 *= wY;
    // sY0 supposed to L( sY0 ) 
    // sV0 supposed to L( sV0 ) 
    Double  m01_pow2  = m01 * m01 ;
    Double  m11_pow2  = m11 * m11 ;
    Double  m21_pow2  = m21 * m21 ;

    Double  temp0 = 2.0 * dU0 * ( m01_pow2 + m11_pow2 + m21_pow2 );
    Double  temp1 = 2.0 *       ( m01_pow2 + m11_pow2 + m21_pow2 );
    Double  temp2 = 2.0 *       ( m01 * ( dY1 - m00 * dY0 - m02 * dV0 - m03 ) + m11 * ( dU1 - m10 * dY0 - m12 * dV0 - m13 ) + m21 * ( dV1 - m20 * dY0 - m22 * dV0 - m23 ) );
/*
    pLinearModel->A += 2.0 * (Double)( sU0*sU0 ) * ( m01_pow2 + m11_pow2 + m21_pow2 );
    pLinearModel->B += 2.0 * (Double)(     sU0 ) * ( m01_pow2 + m11_pow2 + m21_pow2 );
    pLinearModel->E += 2.0 * (Double)(     sU0 ) * ( m01 * ( sY1 - m00 * sY0 - m02 * sV0 - m03 ) + m11 * ( sU1 - m10 * sY0 - m12 * sV0 - m13 ) + m21 * ( sV1 - m20 * sY0 - m22 * sV0 - m23 ) );

    pLinearModel->C += 2.0 * (Double)( sU0     ) * ( m01_pow2 + m11_pow2 + m21_pow2 );
    pLinearModel->D += 2.0 *                       ( m01_pow2 + m11_pow2 + m21_pow2 );
    pLinearModel->F += 2.0 *                       ( m01 * ( sY1 - m00 * sY0 - m02 * sV0 - m03 ) + m11 * ( sU1 - m10 * sY0 - m12 * sV0 - m13 ) + m21 * ( sV1 - m20 * sY0 - m22 * sV0 - m23 ) );
*/
    pLinearModel->A += dU0 * temp0;
    pLinearModel->B += dU0 * temp1;
    pLinearModel->E += dU0 * temp2;

    pLinearModel->C += temp0;
    pLinearModel->D += temp1;
    pLinearModel->F += temp2;
  }
  else if ( iComp == 2 ) { // Cr
    if ( bSeparateLutOptimization ) {
      m02 = 0.0;
      m12 = 0.0;
    }

    dY0 *= wY;
    dY1 *= wY;
    // sY0 supposed to L( sY0 ) 
    // sU0 supposed to L( sU0 ) 
    Double  m02_pow2  = m02 * m02 ;
    Double  m12_pow2  = m12 * m12 ;
    Double  m22_pow2  = m22 * m22 ;

    Double  temp0 = 2.0 * dV0 * ( m02_pow2 + m12_pow2 + m22_pow2 );
    Double  temp1 = 2.0 *       ( m02_pow2 + m12_pow2 + m22_pow2 );
    Double  temp2 = 2.0 *       ( m02 * ( dY1 - m00 * dY0 - m01 * dU0 - m03 ) + m12 * ( dU1 - m10 * dY0 - m11 * dU0 - m13 ) + m22 * ( dV1 - m20 * dY0 - m21 * dU0 - m23 ) );
/*
    pLinearModel->A += 2.0 * (Double)( sV0*sV0 ) * ( m02_pow2 + m12_pow2 + m22_pow2 );
    pLinearModel->B += 2.0 * (Double)(     sV0 ) * ( m02_pow2 + m12_pow2 + m22_pow2 );
    pLinearModel->E += 2.0 * (Double)(     sV0 ) * ( m02 * ( sY1 - m00 * sY0 - m01 * sU0 - m03 ) + m12 * ( sU1 - m10 * sY0 - m11 * sU0 - m13 ) + m22 * ( sV1 - m20 * sY0 - m21 * sU0 - m23 ) );

    pLinearModel->C += 2.0 * (Double)( sV0     ) * ( m02_pow2 + m12_pow2 + m22_pow2 );
    pLinearModel->D += 2.0 *                       ( m02_pow2 + m12_pow2 + m22_pow2 );
    pLinearModel->F += 2.0 *                       ( m02 * ( sY1 - m00 * sY0 - m01 * sU0 - m03 ) + m12 * ( sU1 - m10 * sY0 - m11 * sU0 - m13 ) + m21 * ( sV1 - m20 * sY0 - m21 * sU0 - m23 ) );
*/
    pLinearModel->A += dV0 * temp0;
    pLinearModel->B += dV0 * temp1;
    pLinearModel->E += dV0 * temp2;

    pLinearModel->C += temp0;
    pLinearModel->D += temp1;
    pLinearModel->F += temp2;
  }

  (pLinearModel->nbPts)++;
}

//=======================================================================

TCom1DlutGO::TCom1DlutGO( Int iNbpCode, Int iInputBitDepth, Int iAppendBitDepth, Int iLutBitDepth, Int iSubSampling, Int iBorder, Bool bVerbose )
{
  m_iInputBitDepth  = iInputBitDepth;
  m_iAppendBitDepth = iAppendBitDepth;
  m_iLutBitDepth    = iLutBitDepth;
  m_iSubSampling    = iSubSampling;
  m_iBorder         = iBorder;
  m_bVerbose        = bVerbose;
  m_shift_div       = iNbpCode - 1;

  m_colorData = (ColorData *)malloc( sizeof(ColorData) );
  m_colorData->GainOffsetMatrix     = TComMath::allocateArray2D( 4, 3 );
#if PIECEWISE_LINEAR_1DLUT
  m_colorData->invGainOffsetMatrix  = TComMath::allocateArray2D( 3, 3 );
#endif

  m_nbp       = 1 + ( 1 << ( iNbpCode - 1 ) );

  for ( Int i=0 ; i<3 ; i++ ) {
#if PIECEWISE_LINEAR_1DLUT
    m_pw_lut1D_A[i]     = new TCom1Dlut( iNbpCode, iAppendBitDepth, iAppendBitDepth );
    m_pw_lut1D_B[i]     = new TCom1Dlut( iNbpCode, iAppendBitDepth, iAppendBitDepth );
    m_pw_lut1D_inv_B[i] = new TCom1Dlut( iNbpCode, iAppendBitDepth, iAppendBitDepth );
#endif
  }

  // iModel :
  m_iModel.nbp      = m_nbp;
  m_iModel.iLutLum  = (Int *)malloc( sizeof(Int) * m_nbp );
  m_iModel.iLutCr   = (Int *)malloc( sizeof(Int) * m_nbp );
  m_iModel.iLutCb   = (Int *)malloc( sizeof(Int) * m_nbp );

  m_iModel.nbp_B      = m_nbp;
  m_iModel.iLutLum_B  = (Int *)malloc( sizeof(Int) * m_nbp );
  m_iModel.iLutCr_B   = (Int *)malloc( sizeof(Int) * m_nbp );
  m_iModel.iLutCb_B   = (Int *)malloc( sizeof(Int) * m_nbp );

  razAllData();
}

TCom1DlutGO::~TCom1DlutGO()
{
  TComMath::deleteArray2D( m_colorData->GainOffsetMatrix, 3 );
#if PIECEWISE_LINEAR_1DLUT
  TComMath::deleteArray2D( m_colorData->invGainOffsetMatrix, 3 );
#endif
}

Void  TCom1DlutGO::razData()
{
  for ( Int i=0 ; i<4 ; i++ ) 
  for ( Int j=0 ; j<7 ; j++ ) 
    m_colorData->dtchCoeffMatrix[i][j] = 0.0;

  m_colorData->nbPts = 0;

#if PIECEWISE_LINEAR_1DLUT
  for ( Int i=0 ; i<3 ; i++ ) {
    m_pw_lut1D_A[i]->razData();
    m_pw_lut1D_B[i]->razData();
  }
#endif
}

Void  TCom1DlutGO::razAllData()
{
  razData();

  for ( int i=0 ; i<3 ; i++ )
  for ( int j=0 ; j<4 ; j++ ) {
    m_colorData->GainOffsetMatrix[i][j]     = ( i==j ) ? (1.0) : (0.0) ;
    m_colorData->invGainOffsetMatrix[i][j]  = ( i==j ) ? (1.0) : (0.0) ;
  }
}

Void	TCom1DlutGO::appendColorSamples_GO( TComPicYuv* pcBasePic0, TComPicYuv* pcBasePic1, Bool bLut_B )
{  
  Int stepX = m_iSubSampling;
  Int stepY = m_iSubSampling;

  Int iBorder0   = (m_iBorder >> 1) << 1;
  Int iHeight0   = pcBasePic0->getHeight();
  Int iWidth0    = pcBasePic0->getWidth();
  Int iStride0   = pcBasePic0->getStride();
  Int iCStride0  = pcBasePic0->getCStride();
  Int stepX0     = stepX;
  Int stepY0     = stepY;

  Int iBorder1   = iBorder0;  // b2x ? (iBorder0 << 1) : (iBorder0);
  Int iHeight1   = pcBasePic1->getHeight();
  Int iWidth1    = pcBasePic1->getWidth();
  Int iStride1   = pcBasePic1->getStride();
  Int iCStride1  = pcBasePic1->getCStride();
  Int stepX1     = stepX; // b2x ? (stepX << 1) : (stepX);
  Int stepY1     = stepY; // b2x ? (stepY << 1) : (stepY);

  Pel*	Lum0    = pcBasePic0->getLumaAddr() + iBorder0*iStride0;
  Pel*	Cb0	    = pcBasePic0->getCbAddr()   + (iBorder0>>1)*iCStride0;
  Pel*	Cr0	    = pcBasePic0->getCrAddr()   + (iBorder0>>1)*iCStride0;
  Pel*	Lum1    = pcBasePic1->getLumaAddr() + iBorder1*iStride1;
  Pel*	Cb1	    = pcBasePic1->getCbAddr()   + (iBorder1>>1)*iCStride1;
  Pel*	Cr1	    = pcBasePic1->getCrAddr()   + (iBorder1>>1)*iCStride1;

  Int step_shift_div = 8 - m_shift_div + (m_iInputBitDepth - 8) ;

  for( Int y = iBorder0; y < iHeight0-iBorder0 ; y+=stepY ) {
    for( Int x = iBorder0; x < iWidth0-iBorder0 ; x+=stepX ) {
    
      Pel sY,sU,sV, sY1,sU1,sV1;
      if ( m_iInputBitDepth == m_iAppendBitDepth ) {
        sY = Lum0[x];
        sU = Cb0[x>>1];
        sV = Cr0[x>>1];

        sY1 = Lum1[x];
        sU1 = Cb1[x>>1];
        sV1 = Cr1[x>>1];
      }
      else {
        Int offset  = 1 << ( m_iInputBitDepth - m_iAppendBitDepth - 1 );
        Int shift   = m_iInputBitDepth - m_iAppendBitDepth;

        sY  = ( Lum0[x] + offset )   >> shift ;
        sU  = ( Cb0[x>>1] + offset ) >> shift ;
        sV  = ( Cr0[x>>1] + offset ) >> shift ;

        sY1 = ( Lum1[x] + offset )   >> shift ;
        sU1 = ( Cb1[x>>1] + offset ) >> shift ;
        sV1 = ( Cr1[x>>1] + offset ) >> shift ;
      }
  	
#if PIECEWISE_LINEAR_1DLUT
      sY  = xClip( (Pel)( m_pw_lut1D_A[0]->modelLut( sY, m_iAppendBitDepth, PIECEWISE_LINEAR_1DLUT_CLIPOUT ) + 0.5 ), m_iAppendBitDepth );
      sU  = xClip( (Pel)( m_pw_lut1D_A[1]->modelLut( sU, m_iAppendBitDepth, PIECEWISE_LINEAR_1DLUT_CLIPOUT ) + 0.5 ), m_iAppendBitDepth );
      sV  = xClip( (Pel)( m_pw_lut1D_A[2]->modelLut( sV, m_iAppendBitDepth, PIECEWISE_LINEAR_1DLUT_CLIPOUT ) + 0.5 ), m_iAppendBitDepth );
      if ( bLut_B ) {
        sY1  = xClip( (Pel)( m_pw_lut1D_inv_B[0]->modelLut( sY1, m_iAppendBitDepth, PIECEWISE_LINEAR_1DLUT_CLIPOUT ) + 0.5 ), m_iAppendBitDepth );
        sU1  = xClip( (Pel)( m_pw_lut1D_inv_B[1]->modelLut( sU1, m_iAppendBitDepth, PIECEWISE_LINEAR_1DLUT_CLIPOUT ) + 0.5 ), m_iAppendBitDepth );
        sV1  = xClip( (Pel)( m_pw_lut1D_inv_B[2]->modelLut( sV1, m_iAppendBitDepth, PIECEWISE_LINEAR_1DLUT_CLIPOUT ) + 0.5 ), m_iAppendBitDepth );
      }
#endif
      appendColorData_GO( m_colorData, sY, sU, sV, sY1, sU1, sV1 );
    }
    Lum0  += iStride0 * stepY0;
    Lum1  += iStride1 * stepY1;

    if ( stepY==1 ) {
      if ( y % 2 ) {
        Cb0 += iCStride0;
        Cr0 += iCStride0;
        Cb1 += iCStride1;
        Cr1 += iCStride1;
      }
    }
    else {
      Cb0 += iCStride0 * (stepY0>>1);
      Cr0 += iCStride0 * (stepY0>>1);
      Cb1 += iCStride1 * (stepY1>>1);
      Cr1 += iCStride1 * (stepY1>>1);
    }
  }

  //printf("TComColorMappingAlgo::appendColorSamples() done.\n");
}

Void	TCom1DlutGO::appendColorSamples_1DLut( Int iComp, TComPicYuv* pcBasePic0, TComPicYuv* pcBasePic1, Bool bLut_B, Bool bAppendLut_A )
{
  Int stepX = m_iSubSampling;
  Int stepY = m_iSubSampling;

  Int iBorder0   = (m_iBorder >> 1) << 1;
  Int iHeight0   = pcBasePic0->getHeight();
  Int iWidth0    = pcBasePic0->getWidth();
  Int iStride0   = pcBasePic0->getStride();
  Int iCStride0  = pcBasePic0->getCStride();
  Int stepX0     = stepX;
  Int stepY0     = stepY;

  Int iBorder1   = iBorder0;  // b2x ? (iBorder0 << 1) : (iBorder0);
  Int iHeight1   = pcBasePic1->getHeight();
  Int iWidth1    = pcBasePic1->getWidth();
  Int iStride1   = pcBasePic1->getStride();
  Int iCStride1  = pcBasePic1->getCStride();
  Int stepX1     = stepX; // b2x ? (stepX << 1) : (stepX);
  Int stepY1     = stepY; // b2x ? (stepY << 1) : (stepY);

  Pel*	Lum0    = pcBasePic0->getLumaAddr() + iBorder0*iStride0;
  Pel*	Cb0	    = pcBasePic0->getCbAddr()   + (iBorder0>>1)*iCStride0;
  Pel*	Cr0	    = pcBasePic0->getCrAddr()   + (iBorder0>>1)*iCStride0;
  Pel*	Lum1    = pcBasePic1->getLumaAddr() + iBorder1*iStride1;
  Pel*	Cb1	    = pcBasePic1->getCbAddr()   + (iBorder1>>1)*iCStride1;
  Pel*	Cr1	    = pcBasePic1->getCrAddr()   + (iBorder1>>1)*iCStride1;

  Int step_shift_div = 8 - m_shift_div + (m_iInputBitDepth - 8) ;

  for( Int y = iBorder0; y < iHeight0-iBorder0 ; y+=stepY ) {
    for( Int x = iBorder0; x < iWidth0-iBorder0 ; x+=stepX ) {
    
      Pel sY,sU,sV, sY1,sU1,sV1;
      if ( m_iInputBitDepth == m_iAppendBitDepth ) {
        sY = Lum0[x];
        sU = Cb0[x>>1];
        sV = Cr0[x>>1];

        sY1 = Lum1[x];
        sU1 = Cb1[x>>1];
        sV1 = Cr1[x>>1];
      }
      else {
        Int offset  = 1 << ( m_iInputBitDepth - m_iAppendBitDepth - 1 );
        Int shift   = m_iInputBitDepth - m_iAppendBitDepth;

        sY  = ( Lum0[x] + offset )   >> shift ;
        sU  = ( Cb0[x>>1] + offset ) >> shift ;
        sV  = ( Cr0[x>>1] + offset ) >> shift ;

        sY1 = ( Lum1[x] + offset )   >> shift ;
        sU1 = ( Cb1[x>>1] + offset ) >> shift ;
        sV1 = ( Cr1[x>>1] + offset ) >> shift ;
      }
  	
#if PIECEWISE_LINEAR_1DLUT
      Double  dY1, dU1, dV1;
      if ( bAppendLut_A ) {
        if ( bLut_B ) {
          dY1  = m_pw_lut1D_inv_B[0]->modelLut( sY1, m_iAppendBitDepth, PIECEWISE_LINEAR_1DLUT_CLIPOUT );
          dU1  = m_pw_lut1D_inv_B[1]->modelLut( sU1, m_iAppendBitDepth, PIECEWISE_LINEAR_1DLUT_CLIPOUT );
          dV1  = m_pw_lut1D_inv_B[2]->modelLut( sV1, m_iAppendBitDepth, PIECEWISE_LINEAR_1DLUT_CLIPOUT );
        }
        else {
          dY1  = (Double)sY1;
          dU1  = (Double)sU1;
          dV1  = (Double)sV1;
        }
        modelMatrixInvD( m_colorData->invGainOffsetMatrix, m_colorData->GainOffsetMatrix, dY1, dU1, dV1, &dY1, &dU1, &dV1, m_iAppendBitDepth, m_iAppendBitDepth ); 

        if ( iComp == 0 )       m_pw_lut1D_A[0]->appendColorSample( (Double)sY, dY1 );  // Luma
        else if ( iComp == 1 )  m_pw_lut1D_A[1]->appendColorSample( (Double)sU, dU1 );  // Cb
        else if ( iComp == 2 )  m_pw_lut1D_A[2]->appendColorSample( (Double)sV, dV1 );  // Cr
      }
      else {  // Append lut_B data
        Double  dLY = m_pw_lut1D_A[0]->modelLut( sY, m_iAppendBitDepth, PIECEWISE_LINEAR_1DLUT_CLIPOUT );
        Double  dLU = m_pw_lut1D_A[1]->modelLut( sU, m_iAppendBitDepth, PIECEWISE_LINEAR_1DLUT_CLIPOUT );
        Double  dLV = m_pw_lut1D_A[2]->modelLut( sV, m_iAppendBitDepth, PIECEWISE_LINEAR_1DLUT_CLIPOUT );

        modelMatrixD( m_colorData->GainOffsetMatrix, dLY, dLU, dLV, &dLY, &dLU, &dLV, m_iAppendBitDepth, m_iAppendBitDepth );

        if ( iComp == 0 ) {
          m_pw_lut1D_inv_B[0]->appendColorSample( (Double)sY1, dLY ); // Luma
          m_pw_lut1D_B[0]    ->appendColorSample( dLY, (Double)sY1);  // Luma
        }
        else if ( iComp == 1 ) {
          m_pw_lut1D_inv_B[1]->appendColorSample( (Double)sU1, dLU );  // Cb
          m_pw_lut1D_B[1]    ->appendColorSample( dLU, (Double)sU1 );  // Cb
        }
        else if ( iComp == 2 ) {
          m_pw_lut1D_inv_B[2]->appendColorSample( (Double)sV1, dLV );  // Cr
          m_pw_lut1D_B[2]    ->appendColorSample( dLV, (Double)sV1 );  // Cr
        }
      }
#endif

    }
    Lum0  += iStride0 * stepY0;
    Lum1  += iStride1 * stepY1;

    if ( stepY==1 ) {
      if ( y % 2 ) {
        Cb0 += iCStride0;
        Cr0 += iCStride0;
        Cb1 += iCStride1;
        Cr1 += iCStride1;
      }
    }
    else {
      Cb0 += iCStride0 * (stepY0>>1);
      Cr0 += iCStride0 * (stepY0>>1);
      Cb1 += iCStride1 * (stepY1>>1);
      Cr1 += iCStride1 * (stepY1>>1);
    }
  }

  //printf("TComColorMappingAlgo::appendColorSamples() done.\n");
}

#define xClipMinMax(x,min,max)  ( ((x) < (min)) ? (min) : (((x) > (max)) ? (max) : (x)) )

Void  TCom1DlutGO::iModel_1DLut_GO( Pel Lum0, Pel Cb0, Pel Cr0, Pel* sY1, Pel* sU1, Pel* sV1, Int iInputBitDepth, Int iOuputBitDepth, Bool bLut_B )
{
  Int step_shift_div = 8 - m_shift_div + (iInputBitDepth - 8) ;

  Int iY  = Lum0  >> step_shift_div;
  Int iU  = Cb0   >> step_shift_div;
  Int iV  = Cr0   >> step_shift_div;
    
  iY  = xClipMinMax( iY, 0, m_nbp-1 );
  iU  = xClipMinMax( iU, 0, m_nbp-1 );
  iV  = xClipMinMax( iV, 0, m_nbp-1 );

  Int sY0 = ( Lum0 - (iY<<step_shift_div) ) * m_iModel.iLutLum[iY+1] + ( ((iY+1)<<step_shift_div) - Lum0 ) * m_iModel.iLutLum[iY] ;
  Int sU0 = ( Cb0  - (iU<<step_shift_div) ) * m_iModel.iLutCb[iU+1] +  ( ((iU+1)<<step_shift_div)  - Cb0 ) * m_iModel.iLutCb[iU]  ;
  Int sV0 = ( Cr0  - (iV<<step_shift_div) ) * m_iModel.iLutCr[iV+1] +  ( ((iV+1)<<step_shift_div)  - Cr0 ) * m_iModel.iLutCr[iV]  ;

  Int iShiftGO  = ( m_iModel.iShift_GO + m_iModel.iShift_Lut + ( 8 - m_shift_div ) );
  //Int iOffsetGO = ( iShiftGO > 0 ) ? (1 << (iShiftGO - 1)) : (0);
  Int iOffsetGO = 0;
  assert( "TCom1DlutGO::iModel_1DLut_GO()", iShiftGO >= 0 );

  //Int iShiftGainOffset  = m_iModel.iShift_Lut + ( 8 - m_shift_div );
  Int iShiftGainOffset  = m_iModel.iShift_Lut + ( 8 - m_shift_div ) + (iInputBitDepth - 8);

  Int64 xY1 = ( (m_iModel.iGainOffsetMatrix[0][3]<<iShiftGainOffset) + ( m_iModel.iGainOffsetMatrix[0][0]*sY0 + m_iModel.iGainOffsetMatrix[0][1]*sU0 + m_iModel.iGainOffsetMatrix[0][2]*sV0 ) + iOffsetGO ) ;
  Int64 xU1 = ( (m_iModel.iGainOffsetMatrix[1][3]<<iShiftGainOffset) + ( m_iModel.iGainOffsetMatrix[1][0]*sY0 + m_iModel.iGainOffsetMatrix[1][1]*sU0 + m_iModel.iGainOffsetMatrix[1][2]*sV0 ) + iOffsetGO ) ;
  Int64 xV1 = ( (m_iModel.iGainOffsetMatrix[2][3]<<iShiftGainOffset) + ( m_iModel.iGainOffsetMatrix[2][0]*sY0 + m_iModel.iGainOffsetMatrix[2][1]*sU0 + m_iModel.iGainOffsetMatrix[2][2]*sV0 ) + iOffsetGO ) ;

  if ( bLut_B ) {
    step_shift_div  = iShiftGO + (8 - m_shift_div) + (iInputBitDepth - 8) ;
    iY  = (Int)( xY1  >> step_shift_div );
    iU  = (Int)( xU1  >> step_shift_div );
    iV  = (Int)( xV1  >> step_shift_div );
    
    iY  = xClipMinMax( iY, 0, m_nbp-1 );
    iU  = xClipMinMax( iU, 0, m_nbp-1 );
    iV  = xClipMinMax( iV, 0, m_nbp-1 );

    xY1  = ( xY1 - (iY<<step_shift_div) ) * m_iModel.iLutLum_B[iY+1] + ( ((iY+1)<<step_shift_div) - xY1 ) * m_iModel.iLutLum_B[iY] ;
    xU1  = ( xU1 - (iU<<step_shift_div) ) * m_iModel.iLutCb_B [iU+1] + ( ((iU+1)<<step_shift_div) - xU1 ) * m_iModel.iLutCb_B[iU]  ;
    xV1  = ( xV1 - (iV<<step_shift_div) ) * m_iModel.iLutCr_B [iV+1] + ( ((iV+1)<<step_shift_div) - xV1 ) * m_iModel.iLutCr_B[iV]  ;

    Int iShift  = (iShiftGO + m_iModel.iShift_Lut + (8 - m_shift_div) - (iOuputBitDepth - iInputBitDepth)) ;
    Int iOffset = (1 << (iShift-1)) ;
    //Int iOffset = 0 ;
    (*sY1) = (Pel)( (xY1 + iOffset) >> iShift );
    (*sU1) = (Pel)( (xU1 + iOffset) >> iShift );
    (*sV1) = (Pel)( (xV1 + iOffset) >> iShift );
  }
  else {
    Int iShift  = (iShiftGO - (iOuputBitDepth - iInputBitDepth)) ;
    Int iOffset = (1 << (iShift-1)) ;
    (*sY1) = (Pel)( (xY1 + iOffset) >> iShift );
    (*sU1) = (Pel)( (xU1 + iOffset) >> iShift );
    (*sV1) = (Pel)( (xV1 + iOffset) >> iShift );
  }

  (*sY1) = xClip( (*sY1), iOuputBitDepth );
  (*sU1) = xClip( (*sU1), iOuputBitDepth );
  (*sV1) = xClip( (*sV1), iOuputBitDepth );
}

Void  TCom1DlutGO::compute_iModel( Int iDivShift_Lut, Int iDivShift_GO, Bool bLut_B )
{
  Double  L_Weight  = (Double)( 1 << iDivShift_Lut );
  Double  GO_Weight = (Double)( 1 << iDivShift_GO  );

  m_iModel.iShift_Lut = iDivShift_Lut;
  m_iModel.iShift_GO  = iDivShift_GO;

  for ( Int i=0 ; i<3 ; i++ )
  for ( Int j=0 ; j<4 ; j++ )
    m_iModel.iGainOffsetMatrix[i][j]  = (Int)( 0.5 + GO_Weight * m_colorData->GainOffsetMatrix[i][j] );

#if PIECEWISE_LINEAR_1DLUT
  m_iModel.nbp = m_nbp;
  assert("TCom1DlutGO::compute_iModel():\tbad nbp", m_pw_lut1D_A[0]->getNbPts() == m_nbp );
  assert("TCom1DlutGO::compute_iModel():\tbad nbp", m_pw_lut1D_A[1]->getNbPts() == m_nbp );
  assert("TCom1DlutGO::compute_iModel():\tbad nbp", m_pw_lut1D_A[2]->getNbPts() == m_nbp );
  for ( Int n=0 ; n<m_iModel.nbp ; n++ ) {
      m_iModel.iLutLum[n] = (Int)( m_pw_lut1D_A[0]->getLut(n) * L_Weight + 0.5 );
      m_iModel.iLutCb[n]  = (Int)( m_pw_lut1D_A[1]->getLut(n) * L_Weight + 0.5 );
      m_iModel.iLutCr[n]  = (Int)( m_pw_lut1D_A[2]->getLut(n) * L_Weight + 0.5 );
      m_iModel.iLutLum[n] = (Int)xClip( m_iModel.iLutLum[n], 8+iDivShift_Lut );
      m_iModel.iLutCb[n]  = (Int)xClip( m_iModel.iLutCb[n],  8+iDivShift_Lut );
      m_iModel.iLutCr[n]  = (Int)xClip( m_iModel.iLutCr[n],  8+iDivShift_Lut );
  }
  if ( bLut_B ) {
    m_iModel.nbp_B = m_nbp;
    assert("TCom1DlutGO::compute_iModel():\tbad nbp", m_pw_lut1D_B[0]->getNbPts() == m_nbp );
    assert("TCom1DlutGO::compute_iModel():\tbad nbp", m_pw_lut1D_B[1]->getNbPts() == m_nbp );
    assert("TCom1DlutGO::compute_iModel():\tbad nbp", m_pw_lut1D_B[2]->getNbPts() == m_nbp );
    for ( Int n=0 ; n<m_iModel.nbp_B ; n++ ) {
        m_iModel.iLutLum_B[n] = (Int)( m_pw_lut1D_B[0]->getLut(n) * L_Weight + 0.5 );
        m_iModel.iLutCb_B[n]  = (Int)( m_pw_lut1D_B[1]->getLut(n) * L_Weight + 0.5 );
        m_iModel.iLutCr_B[n]  = (Int)( m_pw_lut1D_B[2]->getLut(n) * L_Weight + 0.5 );
        m_iModel.iLutLum_B[n] = (Int)xClip( m_iModel.iLutLum_B[n], 8+iDivShift_Lut );
        m_iModel.iLutCb_B[n]  = (Int)xClip( m_iModel.iLutCb_B[n],  8+iDivShift_Lut );
        m_iModel.iLutCr_B[n]  = (Int)xClip( m_iModel.iLutCr_B[n],  8+iDivShift_Lut );
    }
  }
#endif
}

Void  TCom1DlutGO::compute1DLutGO( TComPicYuv* pcBasePic0, TComPicYuv* pcBasePic1, Bool bLut_B, Int nbIter )
{
  assert( "TCom1DlutGO::compute1DLutGO():\tBad input images size", pcBasePic0->getWidth() == pcBasePic1->getWidth() );

  // Compute Matrix :
  for ( Int iter=0 ; iter<nbIter ; iter++ ) {

    if ( m_bVerbose ) {
      printf("---------------------\n");
      printf("\tIteration %d\n", iter );
      printf("---------------------\n");
    }

    // GO model :
    if ( true ) {
      razData();

      appendColorSamples_GO( pcBasePic0, pcBasePic1, bLut_B );

      compute_GO_model( m_colorData, m_colorData->GainOffsetMatrix, m_bVerbose );
      if ( !m_colorData->bOk )
        return;

#if PIECEWISE_LINEAR_1DLUT
      // enforce offsets to be zero :
      m_colorData->GainOffsetMatrix[0][3] = 0.0;
      m_colorData->GainOffsetMatrix[1][3] = 0.0;
      m_colorData->GainOffsetMatrix[2][3] = 0.0;
      TComMath::matInverse( (Double **)m_colorData->GainOffsetMatrix, 3, (Double **)m_colorData->invGainOffsetMatrix );
#endif
    }

    // 1D LUT model :
    if ( true && iter<(nbIter-1) )  // last iteration is GO
    for ( Int iter2=0 ; iter2<3 ; iter2++ )
    for ( Int iComp=0 ; iComp<3 ; iComp++ ) {
      razData();

      // 1D Lut A -----------------
      appendColorSamples_1DLut( iComp, pcBasePic0, pcBasePic1, bLut_B );

#if PIECEWISE_LINEAR_1DLUT
      m_pw_lut1D_A[iComp]->computeLut();
      m_pw_lut1D_A[iComp]->displayLut( "1DLUT_A", iComp, m_bVerbose );
#endif

      // 1D Lut B -----------------
      if ( bLut_B ) {
#if PIECEWISE_LINEAR_1DLUT
        // 1D Lut B -----------------
        appendColorSamples_1DLut( iComp, pcBasePic0, pcBasePic1, bLut_B, false );
        m_pw_lut1D_B[iComp]->computeLut();
        m_pw_lut1D_B[iComp]->displayLut( "1DLUT_B", iComp, m_bVerbose );
        m_pw_lut1D_inv_B[iComp]->computeLut();
        m_pw_lut1D_inv_B[iComp]->displayLut( "1DLUT_inv_B", iComp, m_bVerbose );
#endif
      }

    } // iComp

    if ( m_bVerbose ) printf("\n");
  }

  //---------------------------------------------
  //  Display results
  //---------------------------------------------
  if ( !m_bVerbose ) {
    for ( Int iComp=0 ; iComp<3 ; iComp++ )
      m_pw_lut1D_A[iComp]->displayLut( "1DLUT_A", iComp, true );

    printf("\n");
    for ( int i=0 ; i<3 ; i++ )
      printf("\t%g\t%g\t%g\t\t%g\n", m_colorData->GainOffsetMatrix[i][0], m_colorData->GainOffsetMatrix[i][1], m_colorData->GainOffsetMatrix[i][2], m_colorData->GainOffsetMatrix[i][3] );
    printf("\n");

    for ( Int iComp=0 ; iComp<3 ; iComp++ )
      m_pw_lut1D_B[iComp]->displayLut( "1DLUT_B", iComp, true );
  }

  //---------------------------------------------
  //  Compute Integer version
  //---------------------------------------------
#define NB_BITS_STORE 8
  if ( bLut_B ) {
    m_iModel.iShift_Lut  = NB_BITS_STORE - 6;
    m_iModel.iShift_GO   = NB_BITS_STORE + 2;
  }
  else {
    m_iModel.iShift_Lut  = NB_BITS_STORE - 6;
    m_iModel.iShift_GO   = NB_BITS_STORE + 2;
  }

  compute_iModel( m_iModel.iShift_Lut, m_iModel.iShift_GO, bLut_B );
}

Void  TCom1DlutGO::modelMatrix( Double **GainOffsetMatrix, Double sY0, Double sU0, Double sV0, Pel* sY1, Pel* sU1, Pel* sV1, Int InputBitDepth, Int OutputBitDepth )
{
  Double  divGain = (Double)( 1 << (InputBitDepth - 8) );

  Double  dY  = GainOffsetMatrix[0][3]*divGain + ( GainOffsetMatrix[0][0]*sY0 + GainOffsetMatrix[0][1]*sU0 + GainOffsetMatrix[0][2]*sV0 );
  Double  dU  = GainOffsetMatrix[1][3]*divGain + ( GainOffsetMatrix[1][0]*sY0 + GainOffsetMatrix[1][1]*sU0 + GainOffsetMatrix[1][2]*sV0 );
  Double  dV  = GainOffsetMatrix[2][3]*divGain + ( GainOffsetMatrix[2][0]*sY0 + GainOffsetMatrix[2][1]*sU0 + GainOffsetMatrix[2][2]*sV0 );

  Double  multGain = (Double)( 1 << (OutputBitDepth - 8) );

  (*sY1) = xClip( (Pel)( dY * multGain + 0.5 ), OutputBitDepth );
  (*sU1) = xClip( (Pel)( dU * multGain + 0.5 ), OutputBitDepth );
  (*sV1) = xClip( (Pel)( dV * multGain + 0.5 ), OutputBitDepth );
}

Void  TCom1DlutGO::modelMatrixD( Double **GainOffsetMatrix, Double sY0, Double sU0, Double sV0, Double* sY1, Double* sU1, Double* sV1, Int InputBitDepth, Int OutputBitDepth )
{
  Double  divGain = (Double)( 1 << (InputBitDepth - 8) );

  Double  dY  = GainOffsetMatrix[0][3]*divGain + ( GainOffsetMatrix[0][0]*sY0 + GainOffsetMatrix[0][1]*sU0 + GainOffsetMatrix[0][2]*sV0 );
  Double  dU  = GainOffsetMatrix[1][3]*divGain + ( GainOffsetMatrix[1][0]*sY0 + GainOffsetMatrix[1][1]*sU0 + GainOffsetMatrix[1][2]*sV0 );
  Double  dV  = GainOffsetMatrix[2][3]*divGain + ( GainOffsetMatrix[2][0]*sY0 + GainOffsetMatrix[2][1]*sU0 + GainOffsetMatrix[2][2]*sV0 );

  Double  multGain  = (Double)( 1 << (OutputBitDepth - 8) );
  Double  max       = (Double)( (1 << OutputBitDepth) - 1 );

  (*sY1) = xClipD( dY * multGain + 0.5 , max );
  (*sU1) = xClipD( dU * multGain + 0.5 , max );
  (*sV1) = xClipD( dV * multGain + 0.5 , max );
}

Void  TCom1DlutGO::modelMatrixInvD( Double **invGainOffsetMatrix, Double **GainOffsetMatrix, Double sY0, Double sU0, Double sV0, Double* sY1, Double* sU1, Double* sV1, Int InputBitDepth, Int OutputBitDepth )
{
  Double  divGain   = (Double)( 1 << (InputBitDepth - 8) );
  Double  multGain  = (Double)( 1 << (OutputBitDepth - 8) );

  sY0 = (sY0 / multGain) - GainOffsetMatrix[0][3]*divGain;
  sU0 = (sU0 / multGain) - GainOffsetMatrix[1][3]*divGain;
  sV0 = (sV0 / multGain) - GainOffsetMatrix[2][3]*divGain;

  Double  dY  = invGainOffsetMatrix[0][0]*sY0 + invGainOffsetMatrix[0][1]*sU0 + invGainOffsetMatrix[0][2]*sV0 ;
  Double  dU  = invGainOffsetMatrix[1][0]*sY0 + invGainOffsetMatrix[1][1]*sU0 + invGainOffsetMatrix[1][2]*sV0 ;
  Double  dV  = invGainOffsetMatrix[2][0]*sY0 + invGainOffsetMatrix[2][1]*sU0 + invGainOffsetMatrix[2][2]*sV0 ;

  (*sY1) = dY;
  (*sU1) = dU;
  (*sV1) = dV;
}

Void  TCom1DlutGO::colorMap( TComPicYuv* pcInPic, TComPicYuv* pcOutPic, Int outputBitDepth, Bool bLut_B, Bool bUseImodel )
{
  Pel*	Lum0    = pcInPic->getLumaAddr();
  Pel*	Cb0	    = pcInPic->getCbAddr();
  Pel*	Cr0	    = pcInPic->getCrAddr();
  Pel*	Lum1    = pcOutPic->getLumaAddr();
  Pel*	Cb1	    = pcOutPic->getCbAddr();
  Pel*	Cr1	    = pcOutPic->getCrAddr();

  Int step_shift_div = 8 - m_shift_div + (m_iInputBitDepth - 8) ;

  for( Int y = 0; y < pcInPic->getHeight() ; y++ ) {
    for( Int x = 0; x < pcInPic->getWidth() ; x++ ) {
      
      Pel sY1, sU1, sV1;
      if ( !bUseImodel ) {
        // LUT-A :
#if PIECEWISE_LINEAR_1DLUT
        Double  dLY = m_pw_lut1D_A[0]->modelLut( Lum0[x],   m_iInputBitDepth, PIECEWISE_LINEAR_1DLUT_CLIPOUT );
        Double  dLU = m_pw_lut1D_A[1]->modelLut( Cb0[x>>1], m_iInputBitDepth, PIECEWISE_LINEAR_1DLUT_CLIPOUT );
        Double  dLV = m_pw_lut1D_A[2]->modelLut( Cr0[x>>1], m_iInputBitDepth, PIECEWISE_LINEAR_1DLUT_CLIPOUT );
#endif

        if ( bLut_B ) {
          Double  dY1, dU1, dV1;
          modelMatrixD( m_colorData->GainOffsetMatrix, dLY, dLU, dLV, &dY1, &dU1, &dV1, m_iInputBitDepth, m_iInputBitDepth );

#if PIECEWISE_LINEAR_1DLUT
          dLY = m_pw_lut1D_B[0]->modelLut( dY1, m_iInputBitDepth );
          dLU = m_pw_lut1D_B[1]->modelLut( dU1, m_iInputBitDepth );
          dLV = m_pw_lut1D_B[2]->modelLut( dV1, m_iInputBitDepth );

          Double  multGain = (Double)( 1 << (outputBitDepth - m_iInputBitDepth) );

          sY1 = xClip( (Pel)( dLY * multGain + 0.5 ), outputBitDepth );
          sU1 = xClip( (Pel)( dLU * multGain + 0.5 ), outputBitDepth );
          sV1 = xClip( (Pel)( dLV * multGain + 0.5 ), outputBitDepth );
#endif
        }
        else
          modelMatrix( m_colorData->GainOffsetMatrix, dLY, dLU, dLV, &sY1, &sU1, &sV1, m_iInputBitDepth, outputBitDepth );
      }
      else {
        iModel_1DLut_GO( Lum0[x], Cb0[x>>1], Cr0[x>>1], &sY1, &sU1, &sV1, m_iInputBitDepth, outputBitDepth, bLut_B );
      }

      Lum1[x]   = sY1;
      Cb1[x>>1] = sU1;
      Cr1[x>>1] = sV1;
    }
    Lum0  += pcInPic->getStride();
    Lum1  += pcOutPic->getStride();

    if ( y % 2 ) {
      Cb0 += pcInPic->getCStride();
      Cr0 += pcInPic->getCStride();
      Cb1 += pcOutPic->getCStride();
      Cr1 += pcOutPic->getCStride();
    }
  }

}

Void  TCom1DlutGO::createSei( Char* fileName, Bool bLutB )
{
  FILE  *fic = fopen( fileName, "w" );

  Int m_colorMapId          = 0;
  Int m_colorMapCancelFlag  = 0;
  Int m_colorMap_video_signal_type_present_flag = 0;
  Int m_colorMap_video_full_range_flag = 0;
  Int m_colorMap_primaries = 0;
  Int m_colorMap_transfer_characteristics = 0;
  Int m_colorMap_matrix_coeffs = 0;
  Int m_colorMapPersistenceFlag = 0;
  Int m_colorMapModelId = 0;
  Int m_colour_map_coded_data_bit_depth = m_iModel.iShift_Lut + 8;
  Int m_colour_map_target_bit_depth     = m_iModel.iShift_Lut + 8;
  Int m_num_input_pivots[3];
  Int m_num_output_pivots[3];
  Int m_matrix_flag = 1;

  for ( Int i=0 ; i<3 ; i++ ) {
    m_num_input_pivots[i]   = m_iModel.nbp;
    if ( bLutB )
      m_num_output_pivots[i]  = m_iModel.nbp_B;
    else
      m_num_output_pivots[i]  = 0;
  }


  fprintf( fic, "%d\n", m_colorMapId );          // m_colorMapId
  fprintf( fic, "%d\n", m_colorMapCancelFlag );  // m_colorMapCancelFlag
  if ( !m_colorMapCancelFlag ) {
    fprintf( fic, "%d\n", m_colorMapPersistenceFlag );
    fprintf( fic, "%d\n", m_colorMap_video_signal_type_present_flag );
    if ( m_colorMap_video_signal_type_present_flag ) {
      fprintf( fic, "%d\n", m_colorMap_video_full_range_flag );
      fprintf( fic, "%d\n", m_colorMap_primaries );
      fprintf( fic, "%d\n", m_colorMap_transfer_characteristics );
      fprintf( fic, "%d\n", m_colorMap_matrix_coeffs );
      }
  }

  fprintf( fic, "\n");
  fprintf( fic, "%d\n", m_colour_map_coded_data_bit_depth );
  fprintf( fic, "%d\n", m_colour_map_target_bit_depth );
  fprintf( fic, "%d\n", m_colorMapModelId );
  fprintf( fic, "\n");

  assert( "TCom1DlutGO::createSei():\tm_colorMapModelId should be 0", m_colorMapModelId == 0 );


  for ( Int i=0 ; i<3 ; i++ ) {
    fprintf( fic, "%d\n", m_num_input_pivots[i] );
    for ( Int j=0 ; j<m_num_input_pivots[i] ; j++ ) {
      fprintf( fic, "%d ", j << (8-m_shift_div+m_iModel.iShift_Lut) );
      if ( i==0 ) fprintf( fic, "%d ", m_iModel.iLutLum[j] );
      if ( i==1 ) fprintf( fic, "%d ", m_iModel.iLutCb[j] );
      if ( i==2 ) fprintf( fic, "%d ", m_iModel.iLutCr[j] );
    }
    fprintf( fic, "\n");
  }

  fprintf( fic, "\n");
  fprintf( fic, "%d\n", m_matrix_flag );
  if ( m_matrix_flag ) {
    fprintf( fic, "%d\n", m_iModel.iShift_GO );
    for ( Int i=0 ; i<3 ; i++ ) {
      for ( Int j=0 ; j<3 ; j++ )
        fprintf( fic, "%d ", m_iModel.iGainOffsetMatrix[i][j] );
      fprintf( fic, "\n");
    }
  }
  fprintf( fic, "\n");

  for ( Int i=0 ; i<3 ; i++ ) {
    if ( bLutB ) {
      fprintf( fic, "%d\n", m_num_output_pivots[i] );
      for ( Int j=0 ; j<m_num_output_pivots[i] ; j++ ) {
        fprintf( fic, "%d ", j << (8-m_shift_div+m_iModel.iShift_Lut) );
        if ( i==0 ) fprintf( fic, "%d ", m_iModel.iLutLum_B[j] );
        if ( i==1 ) fprintf( fic, "%d ", m_iModel.iLutCb_B[j] );
        if ( i==2 ) fprintf( fic, "%d ", m_iModel.iLutCr_B[j] );
      }
      fprintf( fic, "\n");
    }
    else {
      fprintf( fic, "%d\n", 0 );
    }
  }

  fclose( fic );
}
