/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2014, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __TCOM1DLUTGO__
#define __TCOM1DLUTGO__

// Include files
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "TypeDef.h"
#include "TCom1Dlut.h"


typedef struct {
  Double  a, b;       // Y = a.X + b
  Double  valStart, valEnd;

  Double  A,B,C,D,E,F;

  Int64   sumDc[2], avgDc[2];
  Int64   sumAc[2];

  Int     nbPts;
} linearModel; // node

typedef struct {
  Int iGainOffsetMatrix[3][4];

  Int iShift_Lut;
  Int iShift_GO;
  //Int iDivShift;

  Int* iLutLum;
  Int* iLutCr;
  Int* iLutCb;
  Int nbp;

  Int* iLutLum_B;
  Int* iLutCr_B;
  Int* iLutCb_B;
  Int nbp_B;
} i1DlutGOModel; // node

class TCom1DlutGO
{
  Int m_iInputBitDepth;
  Int m_iAppendBitDepth;
  Int m_iLutBitDepth;
  Int m_iSubSampling;
  Int m_iBorder;
  Int m_shift_div;
  Bool m_bVerbose;

  // Model: 1DLut_B * GO * 1DLut_A
  ColorData*    m_colorData;  // Matrix 3x3
  Int           m_nbp;
  i1DlutGOModel m_iModel;
#if PIECEWISE_LINEAR_1DLUT
  TCom1Dlut*    m_pw_lut1D_A[3];
  TCom1Dlut*    m_pw_lut1D_B[3];
  TCom1Dlut*    m_pw_lut1D_inv_B[3];
#else
  linearModel*  m_lut1D_A[3];
  linearModel*  m_lut1D_B[3];
  //linearModel*  m_lut1D_B_inv[3];
  Pel*          m_lut1D_B_inv_tab[3];
#endif

  Void    razData();
  Void    razAllData();
  inline  Int     xClip( Int x, Int bitDepth );
  inline  Double  xClipD( Double x, Double Max );
 
#if PIECEWISE_LINEAR_1DLUT
#else
  Void  razLut( linearModel** lut1D, Bool bRazDc=true );
  Void  write1DLUT( Char* lutFileName, linearModel* lut1D, Int nbp );

  Void  compute_iModel( linearModel** m_lut1D, Int iDivShift_Lut, Int iDivShift_GO );
  Void  upDate_inverseLutB( linearModel** lut1D, Pel** lut1D_inv_tab, Int comp=-1 );

  Void  display_1DLut( Char *lutName, linearModel** lut1D, Int iComp, Bool bVerbose=m_bVerbose );
  Void  apply_1DLut( linearModel** m_lut1D, Pel sY0, Pel sU0, Pel sV0, Pel* sY1, Pel* sU1, Pel* sV1, Int step_shift_div, Int bitDepth );

  Void  appendColorSamples_and_computeLinearModel_B( TComPicYuv* pcBasePic0, TComPicYuv* pcBasePic1 );

  Void   Model_1DLut_GO( Pel Lum0, Pel Cb0, Pel Cr0, Pel* sY1, Pel* sU1, Pel* sV1, Int iInputBitDepth, Int iOuputBitDepth );
#endif

  Void  iModel_1DLut_GO( Pel Lum0, Pel Cb0, Pel Cr0, Pel* sY1, Pel* sU1, Pel* sV1, Int iInputBitDepth, Int iOuputBitDepth, Bool bLut_B=false );
  Void  compute_iModel( Int iDivShift_Lut, Int iDivShift_GO, Bool bLut_B=false );

public:
  TCom1DlutGO::TCom1DlutGO( Int iNbpCode, Int iInputBitDepth, Int iAppendBitDepth, Int iLutBitDepth, Int iSubSampling, Int iBorder, Bool bVerbose=false );
  TCom1DlutGO::~TCom1DlutGO();

  Void  appendColorSamples_GO( TComPicYuv* pcBasePic0, TComPicYuv* pcBasePic1, Bool bLutB );
  Void  appendColorSamples_1DLut( Int iComp, TComPicYuv* pcBasePic0, TComPicYuv* pcBasePic1, Bool bLut_B, Bool bAppendLut_A=true );
  Void  compute1DLutGO( TComPicYuv* pcBasePic0, TComPicYuv* pcBasePic1, Bool bLut_B, Int nbIter=5 );
  //Void  modelMatrix( Double GainOffsetMatrix[3][4], Double sY0, Double sU0, Double sV0, Pel* sY1, Pel* sU1, Pel* sV1, Int InputBitDepth, Int OutputBitDepth );
  //Void  modelMatrixD( Double GainOffsetMatrix[3][4], Double sY0, Double sU0, Double sV0, Double* sY1, Double* sU1, Double* sV1, Int InputBitDepth, Int OutputBitDepth );
  //Void  modelMatrixInvD( Double invGainOffsetMatrix[3][4], Double sY0, Double sU0, Double sV0, Double* sY1, Double* sU1, Double* sV1, Int InputBitDepth, Int OutputBitDepth );
  Void  modelMatrix( Double **GainOffsetMatrix, Double sY0, Double sU0, Double sV0, Pel* sY1, Pel* sU1, Pel* sV1, Int InputBitDepth, Int OutputBitDepth );
  Void  modelMatrixD( Double **GainOffsetMatrix, Double sY0, Double sU0, Double sV0, Double* sY1, Double* sU1, Double* sV1, Int InputBitDepth, Int OutputBitDepth );
  Void  modelMatrixInvD( Double **invGainOffsetMatrix, Double **GainOffsetMatrix, Double sY0, Double sU0, Double sV0, Double* sY1, Double* sU1, Double* sV1, Int InputBitDepth, Int OutputBitDepth );

  Void  colorMap( TComPicYuv* pcInPic, TComPicYuv* pcOutPic, Int ouputBitDepth, Bool bLut_B=false, Bool bUseImodel=false );

  Void  createSei( Char* fileName, Bool bLutB );
};

inline  Int TCom1DlutGO::xClip( Int x, Int bitDepth )
{ 
  Int Max = (Int)((1<<bitDepth)-1);
//#define Max 255
  Int pel = (Int)( (x < 0) ? 0 : ((x > Max) ? Max : x) );
  return( pel );
}

inline  Double  TCom1DlutGO::xClipD( Double x, Double Max )
{ 
  Double  val = (x < 0.0) ? (0.0) : ((x > Max) ? Max : x) ;
  return( val );
}

#endif