/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2014, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */


// Include files
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <algorithm>    // std::fill
#include "TComColorMappingAlgo.h"
#include "TComLut3D.h"

static  UChar nbpCode[7]={ 0, 2, 3, 5, 9, 17, 33 };

static  Void  assert( Char *comment, Bool bTest )
{
  if ( !bTest ) {
    printf("assertion failed:\t%s\n", comment);
    exit(0);
  }
}

static  Void  static_razColorData( ColorData* pColorData )
{
  for ( Int i=0 ; i<4 ; i++ ) 
  for ( Int j=0 ; j<7 ; j++ ) 
    pColorData->dtchCoeffMatrix[i][j] = 0.0;

  for ( Int i=0 ; i<2 ; i++ ) 
  for ( Int j=0 ; j<2 ; j++ ) 
  for ( Int k=0 ; k<2 ; k++ ) {
    pColorData->nD[i][j][k].dY = 0.0;
    pColorData->nD[i][j][k].dU = 0.0;
    pColorData->nD[i][j][k].dV = 0.0;
  }

  pColorData->sumDiffY  = 0.0;
  pColorData->sumDiffU  = 0.0;
  pColorData->sumDiffV  = 0.0;

  pColorData->nbPts = 0;
}

static  Void  appendColorData( ColorData* pColorData, Pel sY0, Pel sU0, Pel sV0, Pel sY1, Pel sU1, Pel sV1 )
{
  Double  dy_709    = (Double)sY0;
  Double  dcb_709   = (Double)sU0;
  Double  dcr_709   = (Double)sV0;
  Double  dy_2020   = (Double)sY1;
  Double  dcb_2020  = (Double)sU1;
  Double  dcr_2020  = (Double)sV1;

  pColorData->dtchCoeffMatrix[0][0] += dy_709*dy_709;
  pColorData->dtchCoeffMatrix[0][1] += dy_709*dcb_709;
  pColorData->dtchCoeffMatrix[0][2] += dy_709*dcr_709;
  pColorData->dtchCoeffMatrix[0][3] += dy_709;
  pColorData->dtchCoeffMatrix[0][4] += dy_2020*dy_709;
  pColorData->dtchCoeffMatrix[0][5] += dcb_2020*dy_709;
  pColorData->dtchCoeffMatrix[0][6] += dcr_2020*dy_709;

  pColorData->dtchCoeffMatrix[1][0] += dy_709*dcb_709;
  pColorData->dtchCoeffMatrix[1][1] += dcb_709*dcb_709;
  pColorData->dtchCoeffMatrix[1][2] += dcb_709*dcr_709;
  pColorData->dtchCoeffMatrix[1][3] += dcb_709;
  pColorData->dtchCoeffMatrix[1][4] += dy_2020*dcb_709;
  pColorData->dtchCoeffMatrix[1][5] += dcb_2020*dcb_709;
  pColorData->dtchCoeffMatrix[1][6] += dcr_2020*dcb_709;

  pColorData->dtchCoeffMatrix[2][0] += dy_709*dcr_709;
  pColorData->dtchCoeffMatrix[2][1] += dcb_709*dcr_709;
  pColorData->dtchCoeffMatrix[2][2] += dcr_709*dcr_709;
  pColorData->dtchCoeffMatrix[2][3] += dcr_709;
  pColorData->dtchCoeffMatrix[2][4] += dy_2020*dcr_709;
  pColorData->dtchCoeffMatrix[2][5] += dcb_2020*dcr_709;
  pColorData->dtchCoeffMatrix[2][6] += dcr_2020*dcr_709;
  
  pColorData->dtchCoeffMatrix[3][0] += dy_709;
  pColorData->dtchCoeffMatrix[3][1] += dcb_709;
  pColorData->dtchCoeffMatrix[3][2] += dcr_709;
  //pColorData->dtchCoeffMatrix[3][3] = ipic_size;
  (pColorData->dtchCoeffMatrix[3][3])++;
  pColorData->dtchCoeffMatrix[3][4] += dy_2020;
  pColorData->dtchCoeffMatrix[3][5] += dcb_2020;
  pColorData->dtchCoeffMatrix[3][6] += dcr_2020;

  (pColorData->nbPts)++;
}

static  Void  appendColorDataDiff( ColorData* pColorData, Pel sY0, Pel sU0, Pel sV0 )
{
  pColorData->sumDiffY += (Double)sY0;
  pColorData->sumDiffU += (Double)sU0;
  pColorData->sumDiffV += (Double)sV0;

  (pColorData->nbPts)++;
}

// ====================================================================================================================
// Class definition
// ====================================================================================================================
TComColorMappingAlgo::TComColorMappingAlgo( Int iNbNodeCode )
{
  UChar nbpCode[7]      ={ 0, 2, 3, 5, 9, 17, 33 };
  UChar nbpColorData[7] ={ 0, 1, 2, 4, 8, 16, 32 };

  m_nbp       = nbpColorData[ iNbNodeCode ];
  m_shift_div = iNbNodeCode - 1;
  m_colorData = (ColorData *)malloc(sizeof(ColorData)*m_nbp*m_nbp*m_nbp);

  //for ( Int i=0 ; i<(m_nbp*m_nbp*m_nbp) ; i++ )
  //  static_razColorData( &m_colorData[i] );

  razColorData();
}

TComColorMappingAlgo::~TComColorMappingAlgo( )
{
  free(m_colorData);
  m_colorData = NULL;
}

Void  TComColorMappingAlgo::razColorData()
{
  for ( Int i=0 ; i<(m_nbp*m_nbp*m_nbp) ; i++ )
    static_razColorData( &m_colorData[i] );
}

Void	TComColorMappingAlgo::appendColorSamples( TComPicYuv* pcBasePic0,  TComPicYuv* pcBasePic1, Int stepX, Int stepY, Int border, Int InputBitDepth, Bool b2x, Int appendBitDepth )
{
  Int iBorder0   = (border >> 1) << 1;
  Int iHeight0   = pcBasePic0->getHeight();
  Int iWidth0    = pcBasePic0->getWidth();
  Int iStride0   = pcBasePic0->getStride();
  Int iCStride0  = pcBasePic0->getCStride();
  Int stepX0     = stepX;
  Int stepY0     = stepY;

  Int iBorder1   = b2x ? (iBorder0 << 1) : (iBorder0);
  Int iHeight1   = pcBasePic1->getHeight();
  Int iWidth1    = pcBasePic1->getWidth();
  Int iStride1   = pcBasePic1->getStride();
  Int iCStride1  = pcBasePic1->getCStride();
  Int stepX1     = b2x ? (stepX << 1) : (stepX);
  Int stepY1     = b2x ? (stepY << 1) : (stepY);

  Pel*	Lum0    = pcBasePic0->getLumaAddr() + iBorder0*iStride0;
  Pel*	Cb0	    = pcBasePic0->getCbAddr()   + (iBorder0>>1)*iCStride0;
  Pel*	Cr0	    = pcBasePic0->getCrAddr()   + (iBorder0>>1)*iCStride0;
  Pel*	Lum1    = pcBasePic1->getLumaAddr() + iBorder1*iStride1;
  Pel*	Cb1	    = pcBasePic1->getCbAddr()   + (iBorder1>>1)*iCStride1;
  Pel*	Cr1	    = pcBasePic1->getCrAddr()   + (iBorder1>>1)*iCStride1;

  Int step_shift_div = 8 - m_shift_div + (InputBitDepth - 8) ;

  for( Int y = iBorder0; y < iHeight0-iBorder0 ; y+=stepY ) {
    for( Int x = iBorder0; x < iWidth0-iBorder0 ; x+=stepX ) {
    
      Pel sY,sU,sV, sY1,sU1,sV1;
      if ( InputBitDepth == appendBitDepth ) {
        sY = Lum0[x];
        sU = Cb0[x>>1];
        sV = Cr0[x>>1];

        sY1 = Lum1[x];
        sU1 = Cb1[x>>1];
        sV1 = Cr1[x>>1];
      }
      else {
        Int offset  = 1 << ( InputBitDepth - appendBitDepth - 1 );
        Int shift   = InputBitDepth - appendBitDepth;

        sY  = ( Lum0[x] + offset )   >> shift ;
        sU  = ( Cb0[x>>1] + offset ) >> shift ;
        sV  = ( Cr0[x>>1] + offset ) >> shift ;

        sY1 = ( Lum1[x] + offset )   >> shift ;
        sU1 = ( Cb1[x>>1] + offset ) >> shift ;
        sV1 = ( Cr1[x>>1] + offset ) >> shift ;
      }

    	ColorData*  pColorData = getColorData( Lum0[x]>>step_shift_div, Cb0[x>>1]>>step_shift_div, Cr0[x>>1]>>step_shift_div );
    	
      if ( b2x ) {
        sY1 = (Lum1[2*x] + Lum1[2*x+1] + Lum1[2*x+iStride1] + Lum1[2*x+1+iStride1] + 2) >> 2;
        sU1 = (Cb1[x]    + Cb1[x+1]    + Cb1[x+iCStride1]   + Cb1[x+1+iCStride1]   + 2) >> 2;
        sV1 = (Cr1[x]    + Cr1[x+1]    + Cr1[x+iCStride1]   + Cr1[x+1+iCStride1]   + 2) >> 2;
        appendColorData( pColorData, Lum0[x], Cb0[x>>1], Cr0[x>>1], sY1, sU1, sV1 );
      }
      else
        appendColorData( pColorData, sY, sU, sV, sY1, sU1, sV1 );
    }
    Lum0  += iStride0 * stepY0;
    Lum1  += iStride1 * stepY1;
    //if ( (y % 2) ) {
      Cb0 += iCStride0 * (stepY0>>1);
      Cr0 += iCStride0 * (stepY0>>1);
      Cb1 += iCStride1 * (stepY1>>1);
      Cr1 += iCStride1 * (stepY1>>1);
    //}
  }

  //printf("TComColorMappingAlgo::appendColorSamples() done.\n");
}

Void	TComColorMappingAlgo::appendColorSamplesDiff( TComPicYuv* pcBasePic0, TComPicYuv* pcBasePicDiff, Int stepX, Int stepY, Int InputBitDepth )
{
  Int iHeight   = pcBasePic0->getHeight();
  Int iWidth    = pcBasePic0->getWidth();
  Int iStride   = pcBasePic0->getStride();
  Int iCStride  = pcBasePic0->getCStride();

  Pel*	Lum0    = pcBasePic0->getLumaAddr();
  Pel*	Cb0	    = pcBasePic0->getCbAddr();
  Pel*	Cr0	    = pcBasePic0->getCrAddr();
  Pel*	Lum1    = pcBasePicDiff->getLumaAddr();
  Pel*	Cb1	    = pcBasePicDiff->getCbAddr();
  Pel*	Cr1	    = pcBasePicDiff->getCrAddr();

  Int step_shift_div = 8 - m_shift_div + (InputBitDepth - 8) ;

  for( Int y = 0; y < iHeight ; y+=stepY ) {
    for( Int x = 0; x < iWidth ; x+=stepX ) {
    
      Pel sY = Lum0[x];
      Pel sU = Cb0[x>>1];
      Pel sV = Cr0[x>>1];

    	ColorData*  pColorData = getColorData( Lum0[x]>>step_shift_div, Cb0[x>>1]>>step_shift_div, Cr0[x>>1]>>step_shift_div );
    	
      appendColorDataDiff( pColorData, Lum1[x], Cb1[x>>1], Cr1[x>>1] );
    }
    Lum0  += iStride * stepY;
    Lum1  += iStride * stepY;
    //if ( (y % 2) ) {
      Cb0 += iCStride * (stepY>>1);
      Cr0 += iCStride * (stepY>>1);
      Cb1 += iCStride * (stepY>>1);
      Cr1 += iCStride * (stepY>>1);
    //}
  }

  //printf("TComColorMappingAlgo::appendColorSamplesDiff() done.\n");
}

static Double xClipD( Double x , Int bitdepth )
{ 
  //Double  max = (Double)(1 << bitdepth) - 1;
  Double  max = 256.0;

  Double pel = (Pel)( (x < 0.0) ? 0.0 : ( (x > max) ? max : x ) );
  return( pel );
}

//Void  TComColorMappingAlgo::modelMatrix( Double GainOffsetMatrix[3][4], nodeD *nodeIn, nodeD *nodeOut, Int InputBitDepth )
Void  TComColorMappingAlgo::modelMatrix( Double **GainOffsetMatrix, nodeD *nodeIn, nodeD *nodeOut, Int InputBitDepth )
{
  Double  divGain = (Double)( 1 << (InputBitDepth - 8) );

  nodeOut->dY = GainOffsetMatrix[0][3]/divGain + ( GainOffsetMatrix[0][0]*nodeIn->dY + GainOffsetMatrix[0][1]*nodeIn->dU + GainOffsetMatrix[0][2]*nodeIn->dV );
  nodeOut->dU = GainOffsetMatrix[1][3]/divGain + ( GainOffsetMatrix[1][0]*nodeIn->dY + GainOffsetMatrix[1][1]*nodeIn->dU + GainOffsetMatrix[1][2]*nodeIn->dV );
  nodeOut->dV = GainOffsetMatrix[2][3]/divGain + ( GainOffsetMatrix[2][0]*nodeIn->dY + GainOffsetMatrix[2][1]*nodeIn->dU + GainOffsetMatrix[2][2]*nodeIn->dV );
/*
  nodeOut->dY = xClipD( nodeOut->dY, 8 );
  nodeOut->dU = xClipD( nodeOut->dU, 8 );
  nodeOut->dV = xClipD( nodeOut->dV, 8 );
*/
}

Void	TComColorMappingAlgo::compute_GO_model( Int InputBitDepth )
{
  Int step = BOX_MAX >> m_shift_div;
  for ( Int Y=0, i=0 ; Y<BOX_MAX ; Y+=step, i++ )
  for ( Int U=0, j=0 ; U<BOX_MAX ; U+=step, j++ )
  for ( Int V=0, k=0 ; V<BOX_MAX ; V+=step, k++ ) {
   	ColorData*  pColorData = getColorData( i, j, k );
    compute_GO_model( pColorData, pColorData->GainOffsetMatrix );

    if ( m_bVerbose ) {
      printf("GainOffsetMatrix(%d)\t[Y%d][U%d][V%d]:\n", pColorData->nbPts, Y,U,V );
      if ( pColorData->bOk ) {
        for (int row = 0; row<3; row++){
          for (int cl = 0; cl<4; cl++)
              printf("%g\t", pColorData->GainOffsetMatrix [row][cl] );
          printf("\n");
        }
      }
      else
          printf("no model.\n");
    }

    for ( int l=0 ; l<2 ; l++ )
    for ( int m=0 ; m<2 ; m++ )
    for ( int n=0 ; n<2 ; n++ ) {
      nodeD nodeIn;
      
      nodeIn.dY = Y + l*step;
      nodeIn.dU = U + m*step;
      nodeIn.dV = V + n*step;

      modelMatrix( pColorData->GainOffsetMatrix, &nodeIn, &(pColorData->nD[l][m][n]), InputBitDepth );
    }

  }
}

//Void	TComColorMappingAlgo::compute_GO_model( ColorData* pColorData, Double GainOffsetMatrix[3][4] )
Void	TComColorMappingAlgo::compute_GO_model( ColorData* pColorData, Double **GainOffsetMatrix )
{
//eliminationGaussJordan
  int n=4, o=0; 
  double e[5][4]; 

  pColorData->bOk = true;

  for (int row=0; row<=2; row++)
  {
    for (int i=0;i<n;i++){   
      for (int p=0;p<n;p++){ 
        e[p][i] = pColorData->dtchCoeffMatrix [i][p] ; 
        o = p;
      } 
        e[n][i] = pColorData->dtchCoeffMatrix [i][o+row+1] ; 
    } 
    /*********method*************/
 
    // on a saisi les facteurs des equations ds e[][] 
    int y=0; 
    double var1=0,var2=0; 
    double temp; 
    int a,t; 
    for(int x=0;x<n-1;x++){
      for(a=1+x;a<n;a++){ 
        temp=e[x][a]; 
        for (t=x;t<n+1;t++){ 
          e[t][a]=e[t][a]*e[x][x]-e[t][x]*temp; 
          /// triangulation du systeme. 
        } 
      } 
    } 

    if ( e[n-1][n-1] == 0.0 ) {
      if ( m_bVerbose ) printf("Aie.\n");
      pColorData->bOk = false;
      return;
    }

    //Passons au remplacage..  
    GainOffsetMatrix[row][n-1]=e[n][n-1]/e[n-1][n-1]; 
    e[n][n-1]=0; 
    e[n-1][n-1]=0; 
    for (int ligne=1;ligne<=n;ligne++){ 
      for (int sol=2;sol<=n;sol++){ 
        e[n-ligne][n-sol]*=GainOffsetMatrix[row][n-ligne]; 
        e[n][n-sol]-=e[n-ligne][n-sol]; 
        e[n-ligne][n-sol]=0; 
      } 
    if ( n-(ligne+1) >= 0 ) // bug fixing PB
      GainOffsetMatrix[row][n-(ligne+1)]=e[n][n-(ligne+1)]/e[n-(ligne+1)][n-(ligne+1)]; 
    }
  }
   
}

Void  TComColorMappingAlgo::orderingDiff( Int nbFirst, Bool bVerbose, Int *piOrderY, Int *piOrderU, Int *piOrderV, UInt64 *piSumDiffY, UInt64 *piSumDiffU, UInt64 *piSumDiffV )
{
  Bool    bAverage=false;
  Int     nbColorData=m_nbp*m_nbp*m_nbp;
  UInt64  uiSumDiffAll_Y=0, uiSumDiffAll_U=0, uiSumDiffAll_V=0;

  nbFirst = (nbColorData<nbFirst) ? nbColorData : nbFirst ;

  Int* iOrderY  = new Int[nbColorData];
  Int* iOrderU  = new Int[nbColorData];
  Int* iOrderV  = new Int[nbColorData];

  for ( Int i=0 ; i<nbColorData ; i++ ) {
    uiSumDiffAll_Y  +=  (Int)m_colorData[i].sumDiffY;
    uiSumDiffAll_U  +=  (Int)m_colorData[i].sumDiffU;
    uiSumDiffAll_V  +=  (Int)m_colorData[i].sumDiffV;

    if ( bAverage )
      m_colorData[i].sumDiffY = (m_colorData[i].nbPts) ? (m_colorData[i].sumDiffY/(Double)m_colorData[i].nbPts) : (0.0);
    
    iOrderY[i]  = i;
    iOrderU[i]  = i;
    iOrderV[i]  = i;
  }

  for ( Int i=0 ; i<nbFirst ; i++ ) {
    for ( Int j=(i+1) ; j<nbColorData ; j++ ) {
      if ( m_colorData[iOrderY[i]].sumDiffY < m_colorData[iOrderY[j]].sumDiffY ) {
        Int iTmp = iOrderY[i];
        iOrderY[i] = iOrderY[j];
        iOrderY[j] = iTmp;
      }
      if ( m_colorData[iOrderU[i]].sumDiffU < m_colorData[iOrderU[j]].sumDiffU ) {
        Int iTmp = iOrderU[i];
        iOrderU[i] = iOrderU[j];
        iOrderU[j] = iTmp;
      }
      if ( m_colorData[iOrderV[i]].sumDiffV < m_colorData[iOrderV[j]].sumDiffV ) {
        Int iTmp = iOrderV[i];
        iOrderV[i] = iOrderV[j];
        iOrderV[j] = iTmp;
      }
    }
  }

  Int step_shift_div = 8 - m_shift_div;

  for ( Int i=0 ; i<nbFirst ; i++ ) {
    Int sY = iOrderY[i] / (m_nbp*m_nbp);
    Int sU = (iOrderY[i] - sY*(m_nbp*m_nbp)) / m_nbp;
    Int sV = iOrderY[i] - sY*(m_nbp*m_nbp) - sU*m_nbp;

    if ( bVerbose ) {
      if ( bAverage )
        printf("[%d %d %d]\t(%d)\taverage diffY = %g\n",  sY<<step_shift_div, sU<<step_shift_div, sV<<step_shift_div, m_colorData[ iOrderY[i] ].nbPts, m_colorData[ iOrderY[i] ].sumDiffY );
      else
        printf("[%d %d %d]\t(%d)\tsum diffY = %g\n",      sY<<step_shift_div, sU<<step_shift_div, sV<<step_shift_div, m_colorData[ iOrderY[i] ].nbPts, m_colorData[ iOrderY[i] ].sumDiffY );
    }
  }

  (*piOrderY) = iOrderY[nbFirst-1];
  (*piOrderU) = iOrderU[nbFirst-1];
  (*piOrderV) = iOrderV[nbFirst-1];

  delete  iOrderY;
  delete  iOrderU;
  delete  iOrderV;

  if ( piSumDiffY ) (*piSumDiffY) = uiSumDiffAll_Y;
  if ( piSumDiffU ) (*piSumDiffU) = uiSumDiffAll_U;
  if ( piSumDiffV ) (*piSumDiffV) = uiSumDiffAll_V;
}

//=================================================================
//  TComColorMappingMainAlgo
//=================================================================
#define getNbpFromNbpCode(nbpCode)  ( 1 + (1<<(nbpCode-1)) )

TComColorMappingMainAlgo::TComColorMappingMainAlgo( Int iNbpCode, Int iInputBitDepth, Int iAppendBitDepth, Int iLutBitDepth, Int iSubSampling, Int iBorder, Bool bVerbose )
{
  m_verbose       = bVerbose;
  m_nbpCode       = iNbpCode;
  m_nbp           = getNbpFromNbpCode(m_nbpCode);
  m_inputBitDepth = iInputBitDepth;
  m_appendBitDepth= iAppendBitDepth;
  m_lutBitDepth   = iLutBitDepth;
  m_subSampling   = iSubSampling;
  m_border        = iBorder;

  m_nbPtsMinThreshold = 0;

  printf("m_subSampling = %d\n", m_subSampling);

  //assert("subsampling should be even",  m_subSampling == ((m_subSampling>>1)<<1) );
  //assert("border should be even",       m_border      == ((m_border>>1)<<1) );

  pTabColorMapping  = new TComColorMappingAlgo*[m_nbpCode+1];
  pTabLutColor      = new TComLut3D*[m_nbpCode+1];

  for ( Int n=1 ; n<m_nbpCode+1 ; n++ ) {
    pTabColorMapping[n] = new TComColorMappingAlgo( n );
    pTabColorMapping[n]->setVerbose(m_verbose);
    pTabLutColor[n]     = new TComLut3D( getNbpFromNbpCode(n), false );
  }

  m_lutColor  = new TComLut3D( m_nbp, false );
}

TComColorMappingMainAlgo::~TComColorMappingMainAlgo()
{
  for ( Int n=1 ; n<m_nbpCode+1 ; n++ ) {
    delete pTabColorMapping[n];
    delete pTabLutColor[n];
  }
  delete m_lutColor;
}

Void  TComColorMappingMainAlgo::computeCLUT( TComPicYuv* pcBasePic0, TComPicYuv* pcBasePic1 )
{
  for ( Int n=1 ; n<m_nbpCode+1 ; n++ ) {
    pTabColorMapping[n]->razColorData();
    compute_GO_model( pTabColorMapping[n], pcBasePic0, pcBasePic1, n );
    compute_3D_Lut( pTabLutColor[n], pTabColorMapping[n], getNbpFromNbpCode(n), m_lutBitDepth - 8 );
  }

  TComColorMappingAlgo*  pColorMapping  = pTabColorMapping[m_nbpCode];

  Int     denomShift    = m_lutBitDepth - 8;
  Double  d_denomShift  = denomShift ? ((Double)(1<<denomShift)) : (1.0) ;
  Double  d_denomOffset = d_denomShift / 2.0;

  Int step = BOX_MAX / (m_nbp-1);
  for ( Int r=0 ; r<m_nbp ; r++ )
  for ( Int g=0 ; g<m_nbp ; g++ )
  for ( Int b=0 ; b<m_nbp ; b++ ) {
    node* nS = m_lutColor->getNode( r,g,b );
    nS->bUsed = true;

    Double  sumY=0.0;
    Double  sumU=0.0;
    Double  sumV=0.0;
    Int     nbpS=0;
    for ( Int i=0 ; i<2 ; i++ )
    for ( Int j=0 ; j<2 ; j++ )
    for ( Int k=0 ; k<2 ; k++ ) {
      if ( (r-i)>=0 && (r-i)<(m_nbp-1) && (g-j)>=0 && (g-j)<(m_nbp-1) && (b-k)>=0 && (b-k)<(m_nbp-1) ) {
        ColorData *pColorData = pColorMapping->getColorData( r-i, g-j, b-k );
        if ( pColorData->bOk ) {
          nodeD nD = pColorData->nD[i][j][k];
/*
          sumY  += pColorMapping->getNbp() * nD.dY;
          sumU  += pColorMapping->getNbp() * nD.dU;
          sumV  += pColorMapping->getNbp() * nD.dV;
          nbpS  += pColorMapping->getNbp();
*/
          sumY  += pColorData->nbPts * nD.dY;
          sumU  += pColorData->nbPts * nD.dU;
          sumV  += pColorData->nbPts * nD.dV;
          nbpS  += pColorData->nbPts;
        }
      }
    }

    if ( !nbpS )  nS->bUsed = false;

    nbpS  = (nbpS==0) ? (1) : (nbpS);

    if ( !denomShift ) {
      nS->sY = (Short)( 0.5 + (sumY / (Double)nbpS) );
      nS->sU = (Short)( 0.5 + (sumU / (Double)nbpS) );
      nS->sV = (Short)( 0.5 + (sumV / (Double)nbpS) );
    }
    else {
      nS->sY = (Short)( d_denomOffset + d_denomShift*(sumY / (Double)nbpS) );
      nS->sU = (Short)( d_denomOffset + d_denomShift*(sumU / (Double)nbpS) );
      nS->sV = (Short)( d_denomOffset + d_denomShift*(sumV / (Double)nbpS) );
    }

    nS->sY = xClip( nS->sY , 8 + denomShift );
    nS->sU = xClip( nS->sU , 8 + denomShift );
    nS->sV = xClip( nS->sV , 8 + denomShift );

    if ( m_verbose )  printf("nbp%d\t3D lut [%d %d %d] => [%d %d %d]\tnbp=%d\n", m_nbp, r*step, g*step, b*step, nS->sY, nS->sU, nS->sV, nbpS );

  }

  //==============================================================
  //  Checking
  //==============================================================
  for ( Int i=0 ; i<(m_nbp-1) ; i++ )
  for ( Int j=0 ; j<(m_nbp-1) ; j++ )
  for ( Int k=0 ; k<(m_nbp-1) ; k++ ) {
    ColorData *pColorData = pColorMapping->getColorData( i, j, k );
    //if ( pColorData->nbPts ) {
      for ( Int ii=0 ; ii<2 ; ii++ )
      for ( Int jj=0 ; jj<2 ; jj++ )
      for ( Int kk=0 ; kk<2 ; kk++ ) {
        node* nS = m_lutColor->getNode( i+ii, j+jj, k+kk );

        if ( !nS->bUsed || pColorData->nbPts<m_nbPtsMinThreshold ) {
          if ( m_verbose )  printf("############ Issue with [ %d %d %d ]\t(%d)\n", (i+ii)*step, (j+jj)*step, (k+kk)*step, pColorData->nbPts );

          nS->sY = 10000;
          nS->sU = 128;
          nS->sV = 128;

          node  nodeIn;
          nodeIn.sY = (i+ii) * step;
          nodeIn.sU = (j+jj) * step;
          nodeIn.sV = (k+kk) * step;

          //getRecursiveNode( &nodeIn, pTabColorMapping, m_nbpCode-1, nS, m_lutBitDepth-8, m_inputBitDepth );
          getRecursiveNode( &nodeIn, pTabColorMapping, m_nbpCode-1, nS, m_lutBitDepth-8, m_appendBitDepth );

          nS->sY = xClip( nS->sY , 8 + denomShift );
          nS->sU = xClip( nS->sU , 8 + denomShift );
          nS->sV = xClip( nS->sV , 8 + denomShift );

          nS->bUsed = true;

        }
      }
    //}
  }


}

Void  TComColorMappingMainAlgo::compute_3D_Lut( TComLut3D* lutColor, TComColorMappingAlgo* pColorMapping, Int nbp, Int denomShift )
{
  assert("lut size mismatch",  lutColor->getNbp() == nbp );

  Double    d_denomShift  = denomShift ? ((Double)(1<<denomShift)) : (1.0) ;
  Double    d_denomOffset = d_denomShift / 2.0;

  Int         step = BOX_MAX / (nbp-1);
  for ( Int r=0 ; r<nbp ; r++ )
  for ( Int g=0 ; g<nbp ; g++ )
  for ( Int b=0 ; b<nbp ; b++ ) {
    node* n = lutColor->getNode( r,g,b );
    n->bUsed = true;

    Double  sumY=0.0;
    Double  sumU=0.0;
    Double  sumV=0.0;
    Int     nbpS=0;
    for ( Int i=0 ; i<2 ; i++ )
    for ( Int j=0 ; j<2 ; j++ )
    for ( Int k=0 ; k<2 ; k++ ) {
      if ( (r-i)>=0 && (r-i)<(nbp-1) && (g-j)>=0 && (g-j)<(nbp-1) && (b-k)>=0 && (b-k)<(nbp-1) ) {
        ColorData *pColorData = pColorMapping->getColorData( r-i, g-j, b-k );
        if ( pColorData->bOk ) {
          nodeD nD = pColorData->nD[i][j][k];
/*
          sumY  += pColorMapping->getNbp() * nD.dY;
          sumU  += pColorMapping->getNbp() * nD.dU;
          sumV  += pColorMapping->getNbp() * nD.dV;
          nbpS  += pColorMapping->getNbp();
*/
          sumY  += pColorData->nbPts * nD.dY;
          sumU  += pColorData->nbPts * nD.dU;
          sumV  += pColorData->nbPts * nD.dV;
          nbpS  += pColorData->nbPts;
        }
      }
    }

    if ( !nbpS )  n->bUsed = false;

    nbpS  = (nbpS==0) ? (1) : (nbpS);

    if ( !denomShift ) {
      n->sY = (Short)( 0.5 + (sumY / (Double)nbpS) );
      n->sU = (Short)( 0.5 + (sumU / (Double)nbpS) );
      n->sV = (Short)( 0.5 + (sumV / (Double)nbpS) );
    }
    else {
      n->sY = (Short)( d_denomOffset + d_denomShift*(sumY / (Double)nbpS) );
      n->sU = (Short)( d_denomOffset + d_denomShift*(sumU / (Double)nbpS) );
      n->sV = (Short)( d_denomOffset + d_denomShift*(sumV / (Double)nbpS) );
    }

    if ( m_verbose )  printf("nbp%d\t3D lut [%d %d %d] => [%d %d %d]\tnbp=%d\n", nbp, r*step, g*step, b*step, n->sY, n->sU, n->sV, nbpS );
  }

}

Void  TComColorMappingMainAlgo::compute_GO_model( TComColorMappingAlgo* pColorMapping, TComPicYuv* pcBasePic0, TComPicYuv* pcBasePic1, Int nbpCode )
{
  Int nbp = getNbpFromNbpCode( nbpCode );

  Bool  b2x = ( pcBasePic1->getWidth() == 2*pcBasePic0->getWidth() );

  pColorMapping->appendColorSamples( pcBasePic0, pcBasePic1, m_subSampling, m_subSampling, m_border, m_inputBitDepth, b2x, m_appendBitDepth );

  //pColorMapping->compute_GO_model( m_inputBitDepth );
  pColorMapping->compute_GO_model( m_appendBitDepth );
}

Void  TComColorMappingMainAlgo::getRecursiveNode( node* nodeIn, TComColorMappingAlgo** pTabColorMapping, Int nbpCode, node* nodeOut, Int denomShift, Int inputBitDepth )
{
  Int nbp   = getNbpFromNbpCode( m_nbpCode );
  Int step  = BOX_MAX / (nbp-1);

  TComColorMappingAlgo* pColorMapping = pTabColorMapping[nbpCode];

  nodeD nodeDIn, nodeDOut;
  nodeDIn.dY = (Double)nodeIn->sY;
  nodeDIn.dU = (Double)nodeIn->sU;
  nodeDIn.dV = (Double)nodeIn->sV;

  Int i = nodeIn->sY / step;
  Int j = nodeIn->sU / step;
  Int k = nodeIn->sV / step;

  i = ( i >= pColorMapping->getNbp() ) ? (pColorMapping->getNbp()-1) : (i);
  j = ( j >= pColorMapping->getNbp() ) ? (pColorMapping->getNbp()-1) : (j);
  k = ( k >= pColorMapping->getNbp() ) ? (pColorMapping->getNbp()-1) : (k);

  Double    d_denomShift  = denomShift ? ((Double)(1<<denomShift)) : (1.0) ;
  Double    d_denomOffset = d_denomShift / 2.0;

  ColorData *pColorData = pColorMapping->getColorData( i, j, k );
  if ( pColorData->bOk && pColorData->nbPts>=m_nbPtsMinThreshold ) {
    pColorMapping->modelMatrix( pColorData->GainOffsetMatrix, &nodeDIn, &nodeDOut, inputBitDepth );

    if ( !denomShift ) {
      nodeOut->sY = (Short)( nodeDOut.dY + 0.5 );
      nodeOut->sU = (Short)( nodeDOut.dU + 0.5 );
      nodeOut->sV = (Short)( nodeDOut.dV + 0.5 );
    }
    else {
      nodeOut->sY = (Short)( d_denomShift*nodeDOut.dY + d_denomOffset );
      nodeOut->sU = (Short)( d_denomShift*nodeDOut.dU + d_denomOffset );
      nodeOut->sV = (Short)( d_denomShift*nodeDOut.dV + d_denomOffset );
    }

  }
  else {
    if ( nbpCode > 1 )
      getRecursiveNode( nodeIn, pTabColorMapping, nbpCode-1, nodeOut, denomShift, inputBitDepth );
    else {
      if ( m_verbose )
        printf("getRecursiveNode() failed.\n");
      exit(0);
    }
  }

}

static  Void  myFill(Pel *pStart, Pel *pEnd, Pel value)
{
  while ( pStart != pEnd ) {
    *(pStart++) = value;
  }
}

/** Downscale a component array by 2/2 with samples at phase 0 (co-sited) with respect to the high-res grid
 * \returns void
 */
Void TComColorMappingMainAlgo::downScaleComponent2x2( const Pel* pSrc, Pel* pDest, const Int iSrcStride, const Int iDestStride, const Int iSrcWidth, const Int iSrcHeight, const Int inputBitDepth, const Int outputBitDepth )
{
  const int iDestWidth = iSrcWidth/2;
  const int iDestHeight = iSrcHeight/2;

  const int HALF_NUM_DOWN_TAPS = 5;
  const int iTaps[HALF_NUM_DOWN_TAPS] = {40, 0, -12, 0, 4};
  const int iCentreTap = 64;
  int iFilterShift = 7;

  const int iTempStride = iSrcWidth + 2*HALF_NUM_DOWN_TAPS;
  Pel* pTemp = new Pel[ iDestHeight * iTempStride];

  int iTemp;

  // Vertical
  const Pel* pDataSrc = pSrc;
  Pel* pDataDest = pTemp + HALF_NUM_DOWN_TAPS;

  for (int j=0; j<iDestHeight; ++j){
    for (int i=0; i<iSrcWidth; ++i){
      iTemp = pDataSrc[i]*iCentreTap;
      for (int k=0; k<HALF_NUM_DOWN_TAPS; ++k){
        iTemp += iTaps[k]*(pDataSrc[i+(k+1)*iSrcStride] + pDataSrc[i-(k+1)*iSrcStride]);
      }
      iTemp += 1<<(iFilterShift-1);
      //pDataDest[i] = Clip(iTemp>>iFilterShift);
      pDataDest[i] = xClip( iTemp>>iFilterShift, inputBitDepth );
    }
    //std::fill(pDataDest - HALF_NUM_DOWN_TAPS, pDataDest, pDataDest[0] );
    //std::fill(pDataDest + iSrcWidth, pDataDest+iSrcWidth+HALF_NUM_DOWN_TAPS, pDataDest[iSrcWidth-1] );
    myFill(pDataDest - HALF_NUM_DOWN_TAPS, pDataDest, pDataDest[0] );
    myFill(pDataDest + iSrcWidth, pDataDest+iSrcWidth+HALF_NUM_DOWN_TAPS, pDataDest[iSrcWidth-1] );
    pDataSrc += 2*iSrcStride;
    pDataDest += iTempStride;
  }

  iFilterShift += inputBitDepth - outputBitDepth ;

  // Horizontal
  pDataSrc = pTemp+HALF_NUM_DOWN_TAPS;
  pDataDest = pDest;

  for (int j=0; j<iDestHeight; ++j){
    for (int i=0,ii=0; i<iDestWidth; ++i,ii+=2){
      iTemp = pDataSrc[ii]*iCentreTap;
      for (int k=0; k<HALF_NUM_DOWN_TAPS; ++k){
        iTemp += iTaps[k]*(pDataSrc[ii+k+1] + pDataSrc[ii-k-1]);
      }
      iTemp += 1<<(iFilterShift-1);
      //pDataDest[i] = Clip(iTemp>>iFilterShift);
      pDataDest[i] = xClip( iTemp>>iFilterShift, outputBitDepth );

    }
    pDataSrc += iTempStride;
    pDataDest += iDestStride;
  }

  delete[] pTemp;
}

/** Non-normative downscaling of a picture by 2x2, e.g. for original input pictures
 * \returns void
 */
Void TComColorMappingMainAlgo::downScalePic( TComPicYuv* pcYuvSrc, TComPicYuv* pcYuvDest, Int inputBitDepth, Int outputBitDepth )
{
  pcYuvSrc->extendPicBorder();
  downScaleComponent2x2( pcYuvSrc->getLumaAddr(), pcYuvDest->getLumaAddr(), pcYuvSrc->getStride(), pcYuvDest->getStride(), pcYuvSrc->getWidth(), pcYuvSrc->getHeight()   , inputBitDepth, outputBitDepth );
  downScaleComponent2x2( pcYuvSrc->getCbAddr(), pcYuvDest->getCbAddr(), pcYuvSrc->getCStride(), pcYuvDest->getCStride(), pcYuvSrc->getWidth()/2, pcYuvSrc->getHeight()/2 , inputBitDepth, outputBitDepth );
  downScaleComponent2x2( pcYuvSrc->getCrAddr(), pcYuvDest->getCrAddr(), pcYuvSrc->getCStride(), pcYuvDest->getCStride(), pcYuvSrc->getWidth()/2, pcYuvSrc->getHeight()/2 , inputBitDepth, outputBitDepth );
  pcYuvDest->extendPicBorder();
}

