/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2014, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __TCOMLUT3D__
#define __TCOMLUT3D__

#include "TypeDef.h"

#define FIXED_ARRAY     (0)
#define USE_SHIFT_DIV   (1)
#define TETRAHEDRAL     (1)

#if FIXED_ARRAY
  #define NBP 5
#endif
#define BOX_MAX 256

typedef struct {
  Short sY, sU, sV;
  Bool  bEncoded;
  Bool  bUsed;
  Int   iLayer;     // 0: Top layer
} node; // node

typedef struct {
  Bool  bSplit;
  Void  *leave[8];
} octTree;  // octant or cube

typedef struct {
  Bool  bFlag;  // 1: flag, 0: SVLC
  Int   iValue;
} syntaxElement;  // octant or cube

class  TComLut3D
{
  Int     m_nbp;
#if USE_SHIFT_DIV
  Int     m_dr;
  Int     m_dg;
  Int     m_db;
  Int     m_shift_div;  // = log(m_nbp)/log(2)
  Int     m_shift_div3; // = 3*m_shift_div
#endif
#if FIXED_ARRAY
  node    m_yuvLut[NBP][NBP][NBP];
#else
  node    *m_yuvLut;
#endif
  octTree *m_octTree;

  Int     m_inputBitDepth, m_outputBitDepth, m_lutBitDepth;

  Void  interpolLut_fromUpperLayer( node *nodeIn, node *nodeOut, Int iUpperLayer );
  Void  encode(Int iLayer, int r0, int g0, int b0, int errMax, Int errQ, TComLut3D *lutRef, TComLut3D *lutResiduals, octTree *pOctTree, octTree *pOctTreeParent, Bool bVerbose );

  Void    rgb2yuv( node* nIn, node* nOut, Int bd );
  Void    yuv2rgb( node* nIn, node* nOut, Int bd );

public:
  TComLut3D( Int iNbpCode, Bool bAllocateOctree=false );
  ~TComLut3D();

  inline  node    *getNode( int r, int g, int b );
  Void    interpol( node *nodeIn, node *nodeOut );
  Void    interpol_ibd( node *nodeIn, node *nodeOut );
#if TETRAHEDRAL
  Void    interpol_tetra( node *nodeIn, node *nodeOut );
  Void    interpol_tetra_ibd( node *nodeIn, node *nodeOut );
  Void    interpol_tetra_Lum( node *nodeIn, node *nodeOut, Bool bMarkUsedVertices );
  Void    interpol_tetra_Chroma( Short sY, node *nodeIn, node *nodeOut );
#endif
  Void    reset( Bool bResetNode=true, Bool bResetFlag=true );
  Void    setDefault();

  Void    encode( Int errMax, Int errQ, TComLut3D *lutRef, TComLut3D *lutResiduals, Bool bWriteStream );
  Void    decode( Int iLayer, int r0, int g0, int b0, Int errQ, TComLut3D* lutres, octTree *pOctTree, Bool bVerbose );
  Int     getNbpCode();
  Void    prune( Bool b );
  Void    init( TComLut3D* lutRef , Bool bRGB );

  inline Int      getNbp()      { return m_nbp; };
  inline octTree* getOctTree()  { return m_octTree; };

  inline Void setBitDepth( Int inputBitDepth, Int outputBitDepth, Int lutBitDepth );
};

inline  node *TComLut3D::getNode( int r, int g, int b )
{
#if FIXED_ARRAY
  return( &m_yuvLut[r][g][b] );
#else
  return( m_yuvLut + r*m_nbp*m_nbp + g*m_nbp + b );
#endif
}

inline  Void  TComLut3D::setBitDepth( Int inputBitDepth, Int outputBitDepth, Int lutBitDepth )
{
  m_inputBitDepth   = inputBitDepth;
  m_outputBitDepth  = outputBitDepth;
  m_lutBitDepth     = lutBitDepth;
}

#endif

