#include "TComLut3DDouble.h"

TComLut3DDouble::TComLut3DDouble(int nbp)
{
  m_nbp = nbp;
  m_dr  = BOX_MAX / (m_nbp-1);
  m_dg  = BOX_MAX / (m_nbp-1);
  m_db  = BOX_MAX / (m_nbp-1);


  Int nbNode  = m_nbp * m_nbp * m_nbp;
  m_yuvLut    = (nodeD *)malloc(sizeof(nodeD) * nbNode);

  for ( Int n=0 ; n<nbNode ; n++ ) {
    m_yuvLut[n].dY        = 0.0;
    m_yuvLut[n].dU        = 0.0;
    m_yuvLut[n].dV        = 0.0;
    m_yuvLut[n].dY0       = 0.0;
    m_yuvLut[n].dU0       = 0.0;
    m_yuvLut[n].dV0       = 0.0;
    m_yuvLut[n].iNbSample = 0;
    m_yuvLut[n].dist      = m_dr*m_dr + m_dg*m_dg + m_db*m_db;
  }
}

TComLut3DDouble::~TComLut3DDouble()
{
  delete m_yuvLut;
}

Void  TComLut3DDouble::append( Short sY0, Short sU0, Short sV0, Short sY1, Short sU1, Short sV1 )
{
  Int i = sY0 / m_dr;
  Int j = sU0 / m_dg;
  Int k = sV0 / m_db;

  Int er  = sY0 - i*m_dr;
  Int eg  = sU0 - j*m_dg;
  Int eb  = sV0 - k*m_db;

  if ( er > (m_dr - er) ) {
    er  = m_dr - er;
    i++;
  }
  if ( eg > (m_dg - eg) ) {
    eg  = m_dg - eg;
    j++;
  }
  if ( eb > (m_db - eb) ) {
    eb  = m_db - eb;
    k++;
  }

  Int dist  = er*er + eg*eg + eb*eb;

  nodeD*  node  = getNode( i, j, k );

  if ( dist < node->dist ) {
#if 1
    node->dY        = (Double)sY1;
    node->dU        = (Double)sU1;
    node->dV        = (Double)sV1;
#else
    node->dY        = (Double)sY0;
    node->dU        = (Double)sU0;
    node->dV        = (Double)sV0;
#endif
    node->dY0       = (Double)sY0;
    node->dU0       = (Double)sU0;
    node->dV0       = (Double)sV0;
    node->iNbSample = 1;
    node->dist      = dist;

    printf("append (%g;%g;%g) => (%g;%g;%g)\t(dist=%d)\n", node->dY0,node->dU0,node->dV0, node->dY,node->dU,node->dV, node->dist );
  }
  else if ( (dist - node->dist) < 100 ) {
#if 1
    node->dY    +=  (Double)sY1;
    node->dU    +=  (Double)sU1;
    node->dV    +=  (Double)sV1;
#else
    node->dY    +=  (Double)sY0;
    node->dU    +=  (Double)sU0;
    node->dV    +=  (Double)sV0;
#endif
    node->dY0   +=  (Double)sY0;
    node->dU0   +=  (Double)sU0;
    node->dV0   +=  (Double)sV0;
    (node->iNbSample)++;
  }
}

Void  TComLut3DDouble::append( TComPicYuv* pcBasePic, TComPicYuv* pcUsPic )
{
  Int iBWidth   = pcBasePic->getWidth () - pcBasePic->getPicCropLeftOffset() - pcBasePic->getPicCropRightOffset();
  Int iBHeight  = pcBasePic->getHeight() - pcBasePic->getPicCropTopOffset()  - pcBasePic->getPicCropBottomOffset();

  Int iEWidth   = pcUsPic->getWidth () - pcUsPic->getPicCropLeftOffset() - pcUsPic->getPicCropRightOffset();
  Int iEHeight  = pcUsPic->getHeight() - pcUsPic->getPicCropTopOffset()  - pcUsPic->getPicCropBottomOffset();

  Int n=1, d=0;
  if ( iEWidth == 2*iBWidth && iEHeight == 2*iBHeight )           { n = 2;  d = 0; }
  else if ( 2*iEWidth == 3*iBWidth && 2*iEHeight == 3*iBHeight )  { n = 3;  d = 1; }
  else if ( iEWidth == iBWidth && iEHeight == iBHeight )          { n = 1;  d = 0; }
  else
    assert(0);

  Pel*	Lum0  = pcBasePic->getLumaAddr()  + pcBasePic->getPicCropTopOffset()*pcBasePic->getStride()       + pcBasePic->getPicCropLeftOffset();
  Pel*	Cb0	  = pcBasePic->getCbAddr()    + (pcBasePic->getPicCropTopOffset()>>1)*pcBasePic->getCStride() + (pcBasePic->getPicCropLeftOffset()>>1);
  Pel*	Cr0	  = pcBasePic->getCrAddr()    + (pcBasePic->getPicCropTopOffset()>>1)*pcBasePic->getCStride() + (pcBasePic->getPicCropLeftOffset()>>1);

  for( Int y0 = 0; y0 < iBHeight ; y0++ )
  {
    Int y1  = (n * y0) >> (d);
    Pel*	Lum1  = pcUsPic->getLumaAddr()  + (y1+pcUsPic->getPicCropTopOffset())*pcUsPic->getStride()       + pcUsPic->getPicCropLeftOffset();
    Pel*	Cb1	  = pcUsPic->getCbAddr()    + ((y1+pcUsPic->getPicCropTopOffset()>>1))*pcUsPic->getCStride() + (pcUsPic->getPicCropLeftOffset()>>1);
    Pel*	Cr1	  = pcUsPic->getCrAddr()    + ((y1+pcUsPic->getPicCropTopOffset()>>1))*pcUsPic->getCStride() + (pcUsPic->getPicCropLeftOffset()>>1);

    for( Int x0 = 0; x0 < iBHeight ; x0++ )
	  {
      Int x1  = (n * x0) >> (d);

      append( Lum0[x0], Cb0[x0>>1], Cr0[x0>>1], Lum1[x1], Cb1[x1>>1], Cr1[x1>>1] );
    }
    Lum0  += pcBasePic->getStride();
    if ( y0 % 2 ) {
      Cb0   += pcBasePic->getCStride();
      Cr0   += pcBasePic->getCStride();
    }
  }

}

Void  TComLut3DDouble::compute( TComLut3D* outLut3D )
{
  for ( Int i=0 ; i<m_nbp ; i++ )
  for ( Int j=0 ; j<m_nbp ; j++ )
  for ( Int k=0 ; k<m_nbp ; k++ ) {
    //node  *n  = outLut3D->getNode( i, j, k );
    nodeD *nD = getNode( i, j, k );

    assert( nD->iNbSample > 0 );

    if ( nD->iNbSample ) {
      nD->dY  /= (Double)nD->iNbSample;
      nD->dU  /= (Double)nD->iNbSample;
      nD->dV  /= (Double)nD->iNbSample;
      nD->dY0 /= (Double)nD->iNbSample;
      nD->dU0 /= (Double)nD->iNbSample;
      nD->dV0 /= (Double)nD->iNbSample;
      printf("lut3D (%g;%g;%g) => (%g;%g;%g)\t(dist=%d, nb=%d)\n", nD->dY0,nD->dU0,nD->dV0, nD->dY,nD->dU,nD->dV, nD->dist, nD->iNbSample );
      nD->iNbSample = 1;
    }
  }

  eliminationGaussJordan( outLut3D );
}

//
//  equation(i):  e[n][i] = e[0][i]*s[0] + e[1][i]*s[1] + ... + e[n-1][i]*s[n-1]
//  s[]: the solution
Void  TComLut3DDouble::eliminationGaussJordan( TComLut3D* outLut3D )
{
  assert( m_nbp==2 );

  double e[11][10], c_e[11][10]; 
  double s[10]; 

  Double  denom   = m_dr*m_dg*m_db;

  for ( Int comp=0 ; comp<3 ; comp++ ) {

  for ( Int i=0 , eq=0 ; i<m_nbp ; i++ )
  for ( Int j=0 ; j<m_nbp ; j++ )
  for ( Int k=0 ; k<m_nbp ; k++ , eq++ ) {
    nodeD *nD = getNode( i, j, k );

    Double  dY0 = nD->dY0;
    Double  dU0 = nD->dU0;
    Double  dV0 = nD->dV0;

    Int ii = (Int)( dY0 / (Double)m_dr );
    Int jj = (Int)( dU0 / (Double)m_dg );
    Int kk = (Int)( dV0 / (Double)m_db );

    Double  er  = dY0 - (Double)(ii*m_dr);
    Double  eg  = dU0 - (Double)(jj*m_dg);
    Double  eb  = dV0 - (Double)(kk*m_db);

    Double  si[2], sj[2], sk[2];
    si[0] = (m_dr - er);
    si[1] = (er);
    sj[0] = (m_dg - eg);
    sj[1] = (eg);
    sk[0] = (m_db - eb);
    sk[1] = (eb);

    for ( int l=0 , t=0 ; l<2 ; l++ )
    for ( int m=0 ; m<2 ; m++ )
    for ( int n=0 ; n<2 ; n++ , t++ ) {
      c_e[t][eq] = e[t][eq] = si[l]*sj[m]*sk[n] / denom;
    }

    if ( comp == 0 )  e[8][eq] = nD->dY;
    if ( comp == 1 )  e[8][eq] = nD->dU;
    if ( comp == 2 )  e[8][eq] = nD->dV;

    c_e[8][eq] = e[8][eq];
  }

  int n=8; 

  // on a saisi les facteurs des equations ds e[][] 
  //int y=0; 
  //double var1=0,var2=0; 
  double temp; 
  int a,t; 
  for(int x=0;x<n-1;x++){ 
    for(a=1+x;a<n;a++){ 
      temp=e[x][a]; 
      for (t=x;t<n+1;t++){ 
        e[t][a]=e[t][a]*e[x][x]-e[t][x]*temp; 

        /// triangulation du systeme. 
      } 
    } 
  } 

  //Passons au remplacage.. 
//assert( n==8 );
//assert( e[n-1][n-1] != 0.0 );
  s[n-1]=e[n][n-1]/e[n-1][n-1]; 
  e[n][n-1]=0; 
  e[n-1][n-1]=0; 
  for (int ligne=1;ligne<=n;ligne++){ 
    for (int sol=2;sol<=n;sol++){ 
      e[n-ligne][n-sol]*=s[n-ligne]; 
      e[n][n-sol]-=e[n-ligne][n-sol]; 
      e[n-ligne][n-sol]=0; 
    } 
//assert( n-(ligne+1) >= 0 );
//assert( e[n-(ligne+1)][n-(ligne+1)] != 0.0 );
    if ( n-(ligne+1) >= 0 ) // bug fixing PB
      s[n-(ligne+1)]=e[n][n-(ligne+1)]/e[n-(ligne+1)][n-(ligne+1)]; 
  } 

  // Resolution.. 

  // Affichage.. 

  for (int af=0;af<n;af++){
    cout<<" var "<<af<<" = "<<s[af];
  } 
  cout<<"\n"; 
/*
  // checking :
  for ( Int eq=0 ; eq<n ; eq++ ) {
    Double  sum=0.0;
    for ( int j=0 ; j<n ; j++ )
      sum += c_e[j][eq]*s[j];
    printf("check(%d,%d):\t%g = %g ?\n", comp,eq, c_e[8][eq], sum );
  }
*/

  for ( Int i=0 , tt=0 ; i<m_nbp ; i++ )
  for ( Int j=0 ; j<m_nbp ; j++ )
  for ( Int k=0 ; k<m_nbp ; k++ , tt++ ) {
    node  *nn= outLut3D->getNode( i, j, k );

    if ( comp == 0 )  nn->sY = (Short)( s[tt] + 0.5 );
    if ( comp == 1 )  nn->sU = (Short)( s[tt] + 0.5 );
    if ( comp == 2 )  nn->sV = (Short)( s[tt] + 0.5 );

    if ( comp == 2 )
      printf("lut3D v2\t(%d;%d;%d)\n", nn->sY, nn->sU, nn->sV );
  }

  } // for comp

  
  
  for ( Int nn=0 ; nn<(m_nbp*m_nbp*m_nbp) ; nn++ ) {
    node  nodeIn, nodeOut;
    nodeD *nD = &(m_yuvLut[nn]);

    nodeIn.sY = (Short)( nD->dY0 + 0.5 );
    nodeIn.sU = (Short)( nD->dU0 + 0.5 );
    nodeIn.sV = (Short)( nD->dV0 + 0.5 );
    outLut3D->interpol( &nodeIn, &nodeOut );
    printf("check:\t%d/%d %d/%d %d/%d\n", 
      nodeOut.sY, (Int)( nD->dY + 0.5 ),
      nodeOut.sU, (Int)( nD->dU + 0.5 ),
      nodeOut.sV, (Int)( nD->dV + 0.5 ) );
  }

}

