// Include files
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <math.h>
#include "linalg.h"   // ALGLIB
#include "solvers.h"  // ALGLIB
#include "TComMath.h"

using namespace alglib;

#define USE_ALGLIB  (1)

Double** TComMath::allocateArray2D( Int sx, Int sy )
{
  Double  **mat = (Double **)malloc( sizeof(Double *) * sy );
  Double  *tmp  = (Double *) malloc( sizeof(Double) * sx * sy );

  for ( Int y=0 ; y<sy ; y++ )
    mat[y] = &tmp[y*sx];
  //mat[y] = (Double *)malloc( sizeof(Double) * sx );

  return mat ;
}

Void TComMath::deleteArray2D( Double **mat, Int sy )
{
/*
  for ( Int y=0 ; y<sy ; y++ )
    free( mat[y] );
*/
  free( mat[0] );
  free( mat );
}

Void TComMath::displayArray2D( Char *matName, Double **mat, Int sx, Int sy )
{
  printf("%s\n", matName );
  for ( int i=0 ; i<sy ; i++ ) {
    for ( int j=0 ; j<sx ; j++ )
      printf("\t%g", mat[i][j] );
    printf("\n");
  }
}

/*
   Recursive definition of determinate using expansion by minors.
*/
Double TComMath::matDeterminant( Double **a, Int n )
{
   Int i,j,j1,j2;
   Double det = 0;
   Double **m = NULL;

   if (n < 1) { /* Error */

   } else if (n == 1) { /* Shouldn't get used */
      det = a[0][0];
   } else if (n == 2) {
      det = a[0][0] * a[1][1] - a[1][0] * a[0][1];
   } else {
      det = 0;
      for (j1=0;j1<n;j1++) {
         m = (Double **)malloc((n-1)*sizeof(Double *));
         for (i=0;i<n-1;i++)
            m[i] = (Double *)malloc((n-1)*sizeof(Double));
         for (i=1;i<n;i++) {
            j2 = 0;
            for (j=0;j<n;j++) {
               if (j == j1)
                  continue;
               m[i-1][j2] = a[i][j];
               j2++;
            }
         }
         det += pow(-1.0,j1+2.0) * a[0][j1] * matDeterminant(m,n-1);
         for (i=0;i<n-1;i++)
            free(m[i]);
         free(m);
      }
   }
   return(det);
}

/*
   Find the cofactor matrix of a square matrix
*/
Void TComMath::matCoFactor( Double **a, Int n, Double **b )
{
   int i,j,ii,jj,i1,j1;
   double det;
   double **c;

   c = (Double **)malloc((n-1)*sizeof(double *));
   for (i=0;i<n-1;i++)
     c[i] = (Double *)malloc((n-1)*sizeof(double));

   for (j=0;j<n;j++) {
      for (i=0;i<n;i++) {

         /* Form the adjoint a_ij */
         i1 = 0;
         for (ii=0;ii<n;ii++) {
            if (ii == i)
               continue;
            j1 = 0;
            for (jj=0;jj<n;jj++) {
               if (jj == j)
                  continue;
               c[i1][j1] = a[ii][jj];
               j1++;
            }
            i1++;
         }

         /* Calculate the determinate */
         det = matDeterminant(c,n-1);

         /* Fill in the elements of the cofactor */
         b[i][j] = pow(-1.0,i+j+2.0) * det;
      }
   }
   for (i=0;i<n-1;i++)
      free(c[i]);
   free(c);
}

/*
   Transpose of a square matrix, do it in place
*/
Void TComMath::matTranspose( Double **a, Int n )
{
   int i,j;
   double tmp;

   for (i=1;i<n;i++) {
      for (j=0;j<i;j++) {
         tmp = a[i][j];
         a[i][j] = a[j][i];
         a[j][i] = tmp;
      }
   }
}

Bool TComMath::matSolve( Double **mat, Double *Y, Int nbp, Double *X )
{
#if USE_ALGLIB
  real_2d_array ra;
  real_1d_array rb, rx;

  ra.setcontent( nbp, nbp, mat[0] );
  rb.setcontent( nbp, Y );

  ae_int_t          info;
  densesolverreport rep;
  rmatrixsolve( ra, nbp, rb, info, rep, rx );

  for ( Int i=0 ; i<nbp ; i++ ) X[i] = rx[i];

  return (info == 1);
#else
  Double  **mat_inv = allocateArray2D( nbp, nbp );

  matInverse( mat, nbp, mat_inv );

  for ( Int i=0 ; i<nbp ; i++ ) {
    X[i] = 0.0;
    for ( Int j=0 ; j<nbp ; j++ )
      X[i]  += mat_inv[i][j] * Y[j];
  }
  deleteArray2D( mat_inv, nbp );

  return true;
#endif
}

Void  TComMath::matInverse( Double **a, Int n, Double **b )
{
#if USE_ALGLIB
  copyArray2D( a, b, n );

  real_2d_array rb;
  rb.setcontent( n, n, b[0] );

  ae_int_t      info;
  matinvreport  rep;
  rmatrixinverse( rb, info, rep );

  for ( Int i=0 ; i<n ; i++ )
  for ( Int j=0 ; j<n ; j++ )
    b[i][j] = rb[i][j];
#else
  Double  det = matDeterminant( a, n );
  matCoFactor( a, n, b );
  matTranspose( b, n );
  matMult( b, det, n, n );
#endif
/*
  Double **Id = allocateArray2D( n, n );
  matMult( a, b, Id, n );
  displayArray2D( "Check matrix Identity", Id, n, n );
  deleteArray2D( Id, n );
*/
}


