/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2014, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef __TCOMMATH__
#define __TCOMMATH__

// Include files
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "TypeDef.h"


class TComMath
{
public:
  static  Double**  allocateArray2D( Int sx, Int sy );
  static  Void      deleteArray2D( Double **mat, Int sy );
  static  Void      copyArray2D( Double **src, Double **dst, Int sy );
  static  Void      displayArray2D( Char *matName, Double **mat, Int sx, Int sy );

  static          Double  matDeterminant( Double **a, Int n );
  static          Void    matCoFactor( Double **a, Int n, Double **b);
  static          Void    matTranspose( Double **a, Int n );
  static          Void    matMult( Double **a, Double K, Int sx, Int sy );
  static          Void    matMult( Double **a, Double **b, Double **c, Int n );
  static  inline  Void    arrayMult( Double *a, Double K, Int sx );
  static          Void    matInverse( Double **a, Int n, Double **b );
  static          Bool    matSolve( Double **mat, Double *Y, Int nbp, Double *X );

  static  inline  Pel     xClip( Pel x, Int bitDepth );
  static  inline  Double  xClipD( Double x, Double Max );
};

inline  Void  TComMath::matMult( Double **a, Double K, Int sx, Int sy )
{
  for ( Int i=0 ; i<sy ; i++ )
  for ( Int j=0 ; j<sx ; j++ )
    a[i][j] /= K ;
}

inline  Void  TComMath::arrayMult( Double *a, Double K, Int sx )
{
  for ( Int j=0 ; j<sx ; j++ )
    a[j] /= K ;
}

inline  Void  TComMath::matMult( Double **a, Double **b, Double **c, Int n )
{
  for ( Int i=0 ; i<n ; i++ )
    for ( Int j=0 ; j<n ; j++ ) {
      c[i][j] = 0.0 ;
      for ( Int k=0 ; k<n ; k++ )  
        c[i][j] += a[i][k] * b[k][j];
    }
}

inline  Pel TComMath::xClip( Pel x, Int bitDepth )
{ 
  Int Max = (Int)((1<<bitDepth)-1);
//#define Max 255
  Pel pel = (Pel)( (x < 0) ? 0 : ((x > Max) ? Max : x) );
  return( pel );
}

inline  Double  TComMath::xClipD( Double x, Double Max )
{ 
  Double  val = (x < 0.0) ? (0.0) : ((x > Max) ? Max : x) ;
  return( val );
}

inline  Void  TComMath::copyArray2D( Double **src, Double **dst, Int n )
{
  for ( Int i=0 ; i<n ; i++ )
    for ( Int j=0 ; j<n ; j++ )
      dst[i][j] = src[i][j] ;
}

#endif  // __TCOMMATH__
