/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2012, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TComToneMapping.h
    \brief    Tone Mapping class (header)
*/

#ifndef __TCOMTONEMAPPING__
#define __TCOMTONEMAPPING__


// Include files
#include "TComPicYuv.h"
#include "TComLut3D.h"

#define MODEL_MATRIX    (0)
#define MODEL_3DLUT     (1)
#define MODEL_3DLUT_v2  (2)

// ====================================================================================================================
// Class definition
// ====================================================================================================================
/// tone mapping class
class TComToneMapping
{
  Bool        m_bVerbose;
  Char*       m_toneMappingFile;
  Int         m_model;            // MODEL_MATRIX or MODEL_3DLUT
  Int         m_nbpCode;
	TComPicYuv* m_pcTempPic;
  TComLut3D*  m_lutRef;
  TComLut3D*  m_lutCoded;
  TComLut3D*  m_lutResiduals;
  Int         m_errQ;

  Int         **m_gain;
  Int         *m_offset;
  Int         m_denom;
  Int         m_denom_shift;

  Int         m_inputBitDepth, m_outputBitDepth, m_lutBitDepth;

  Void  loadFile( Char *cFile );
  Void  modelMatrix( Int** matrix, Int* offset, Int denom, node *nodeIn, node *nodeOut );

  inline  Pel xClip( Pel x, Int ibd=8 );

public:
  TComToneMapping( Int iModel, Int iNbpCode, Char *cFile );
  ~TComToneMapping();

  Void        decode();

  Void                writeFile( Char *cFile, TComLut3D* lut=NULL );
  Void                dumpPicture( Char* fileName, TComPicYuv*	pPic, Int iShift );
  Void                toneMap( TComPicYuv* pcBasePic, TComLut3D *lut, Bool bMarkUsedVertices=false );
  inline Int          getNbpCode()      { return m_nbpCode;};
  inline TComPicYuv*  getToneMap()      { return m_pcTempPic;};
  inline Void         setNbpCode(Int n) { m_nbpCode = n;};
  inline Int          getNbp();

  inline Int        getModel()          { return m_model; };
  inline Void       setModel( Int im )  { m_model = im; };
  inline TComLut3D* getLutResiduals()   { return m_lutResiduals; };
  inline TComLut3D* getLutRef()         { return m_lutRef; };
  inline octTree*   getOctTree()        { return m_lutCoded->getOctTree(); };
  inline Int        getErrQ()           { return m_errQ; };
  inline Void       setErrQ( Int e )    { m_errQ = e; };

  inline Int**      getMatrixGain()         { return (Int**)m_gain; };
  inline Int*       getMatrixOffset()       { return (Int*)m_offset; };
  inline Int        getMatrixDenom()        { return m_denom; };
  inline Void       setMatrixDenom( Int d ) { m_denom = d; };

  inline Void       setBitDepth( Int inputBitDepth, Int outputBitDepth, Int lutBitDepth );
};

inline  Pel TComToneMapping::xClip( Pel x, Int bitDepth )
{ 
  Int Max = (Int)((1<<bitDepth)-1);
//#define Max 255
  Pel pel = (Pel)( (x < 0) ? 0 : ((x > Max) ? Max : x) );
  return( pel );
}

inline  Void  TComToneMapping::setBitDepth( Int inputBitDepth, Int outputBitDepth, Int lutBitDepth )
{
  m_inputBitDepth   = inputBitDepth;
  m_outputBitDepth  = outputBitDepth;
  m_lutBitDepth     = lutBitDepth;

  if ( m_lutCoded ) m_lutCoded->setBitDepth( m_inputBitDepth, m_outputBitDepth, m_lutBitDepth );
}

#endif 
