#include <windows.h>
#include <stdio.h>
#include <string.h>	// memcpy
#include "TR_perfTime.h"


bool				PerfTime::m_initDone=false;
LARGE_INTEGER		PerfTime::m_startAbsoluteTime;
CRITICAL_SECTION	PerfTime::m_monitor;

PerfTime::PerfTime()
{
	int	ret;
	
	if ( !m_initDone ) {
		InitializeCriticalSection(&m_monitor);

		ret = QueryPerformanceCounter(&m_startAbsoluteTime);

		m_initDone = true;
	}

	m_intMax32bits = (unsigned int)0xFFFFFFFF;
	
	if ( !(ret=QueryPerformanceFrequency(&m_freqTime)) ) {
		printf("PerfTime::PerfTime():QueryPerformanceFrequency() High resolution conter not supported\n");
		exit(0);
	}

	if ( m_freqTime.HighPart ) {
		printf("PerfTime::PerfTime():QueryPerformanceFrequency() returned %d %d\n", m_freqTime.HighPart, m_freqTime.LowPart);
		exit(0);
	}
	
	ret = QueryPerformanceCounter(&m_startTime);
	memcpy(&m_lastTime, &m_startTime, sizeof(LARGE_INTEGER));
}

int	PerfTime::diffTime(LARGE_INTEGER *t2, LARGE_INTEGER *t1)
{
//	Returns the time elapsed in ms
/*
	unsigned int	LowPart, HighPart;
	double			dLow, dHigh;

	
	if ( t2->LowPart < t1->LowPart ) {
		LowPart		= t2->LowPart + (m_intMax32bits - t1->LowPart);
		HighPart	= t2->HighPart - t1->HighPart - 1;
	}
	else {
		LowPart		= t2->LowPart - t1->LowPart;
		HighPart	= t2->HighPart - t1->HighPart;
	}
		
	printf("PerfTime::diffTime():\tHighPart=%d\tLowPart=%d\t(Freq = %d %d)\n", HighPart, LowPart, m_freqTime.HighPart, m_freqTime.LowPart);

	printf("PerfTime::diffTime():\tt2(%I64d)-t1(%I64d)=%I64d\t(freqTime.QuadPart=%I64d)\n",
		t2->QuadPart, t1->QuadPart, t2->QuadPart - t1->QuadPart, m_freqTime.QuadPart);

	dLow	= (double)LowPart/(double)(m_freqTime.LowPart);
	dHigh	= (double)HighPart*((double)m_intMax32bits/(double)(m_freqTime.LowPart));

	printf("PerfTime::diffTime():\tdLow=%g\tdHigh=%g\t(freqTime.LowPart=%d)\tm_freqTime=%d\n", dLow, dHigh, m_freqTime.LowPart, &m_freqTime);

	if ( !HighPart )
		return( (int)( 1000.0*dLow + 0.5 ) );
	else
		return( (int)( 1000.0*(dHigh + dLow) + 0.5 ) );
*/

	__int64	deltaQuadPart	= (t2->QuadPart - t1->QuadPart);
/* -------- OLD CODE : -----------------------
	int		ifreq			= (int)((m_freqTime.QuadPart/1000000) & 0xFFFFFFFF);
	int		idelta			= (int)((deltaQuadPart/1000) & 0xFFFFFFFF);

	int		ratio			= idelta / ifreq;

	if ( ratio <= 0 )
		printf("PerfTime::diffTime():\tidelta=%d\tifreq=%d\tratio=%d\tdeltaQuadPart=%I64d\n", idelta, ifreq, ratio, deltaQuadPart);
   -------- END OF OLD CODE : ----------------------- */
	__int64		ifreq			= m_freqTime.QuadPart / 1000000;
	__int64		idelta			=  deltaQuadPart / 1000;

	int		ratio			= (int)( idelta / ifreq );

	if ( ratio <= 0 )
		printf("PerfTime::diffTime():\tidelta=%I64d\tifreq=%I64d\tratio=%d\tdeltaQuadPart=%I64d\n", idelta, ifreq, ratio, deltaQuadPart);

	return( ratio );
}

int	PerfTime::elapsed()
{
//	Returns the time elapsed in ms
	LARGE_INTEGER	m_curTime;
	int				ret;
	
	ret = QueryPerformanceCounter(&m_curTime);

	ret = diffTime(&m_curTime, &m_lastTime);
		
	//printf("PerfTime::elapsed():\tret=%d\n", ret);

	memcpy(&m_lastTime, &m_curTime, sizeof(LARGE_INTEGER));

	return(ret);
}

int	PerfTime::getAbsoluteTime()
{
//	Returns the time elapsed in ms
	LARGE_INTEGER	m_curTime;
	int				ret;

	
	ret = QueryPerformanceCounter(&m_curTime);

	EnterCriticalSection(&m_monitor);
	
		ret = diffTime(&m_curTime, &m_startAbsoluteTime);

	LeaveCriticalSection(&m_monitor);

	return(ret);
}

int	PerfTime::getTime()
{
//	Returns the time elapsed in ms
	LARGE_INTEGER	m_curTime;
	int				ret;

	
	ret = QueryPerformanceCounter(&m_curTime);
	
	ret = diffTime(&m_curTime, &m_startTime);

	return(ret);
}

PerfTime::~PerfTime()
{
}
