#ifndef _PerfTime_
#define _PerfTime_

class PerfTime {
/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2014, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
	//----------------------------------------------------
	//	All the times are expessed in millisec.
	//----------------------------------------------------
	private:
		static	LARGE_INTEGER		m_startAbsoluteTime;	// very first date initialized at the very first PerfTime object creation
		static	bool				m_initDone;				// true the very first time, false else
		static	CRITICAL_SECTION	m_monitor;				// used to allow multiple PerfTime instance to access to "m_startAbsoluteTime"

		LARGE_INTEGER			m_startTime;		// creation date of this instance
		LARGE_INTEGER			m_lastTime;			// date of last call to "elapsed()"
		LARGE_INTEGER			m_freqTime;			// timer tick frequency
		unsigned int			m_intMax32bits;		// max value of a 32 bits integer
		
	public:
		PerfTime();
		~PerfTime();
		
		int	elapsed();				// returns the time in ms elapsed since the previous call
		int	getAbsoluteTime();		// returns the time in ms elapsed since very first PerfTime instance creation
		int	getTime();				// returns the time in ms elapsed since this PerfTime instance has been created

	private:
		int	diffTime(LARGE_INTEGER *t2, LARGE_INTEGER *t1);
		
};

#endif