#include <windows.h>  // TR_PerfTime...

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "TypeDef.h"
#include "Param.h"
#include "misc.h"

#include "TComToneMapping.h"
#include "TComColorMappingAlgo.h"
#include "TComLut3D.h"
#include "TR_PerfTime.h"
#include "TCom1DlutGO.h"


Void  buildNameSequence( Char *rootName, Char *ext, Int sizeX, Int sizeY, Int frameRate , Char *nameOut, Int bitDepth=8 )
{
  if ( !ext )
    sprintf( nameOut, "%s_%dx%d_%dbit_%d.yuv", rootName, sizeX, sizeY, bitDepth, frameRate );
  else
    sprintf( nameOut, "%s_%s_%dx%d_%dbit_%d.yuv", rootName, ext, sizeX, sizeY, bitDepth, frameRate );
}


Int getNb( Int iNbCode )
{
  UChar nbpCode[7]={ 0, 2, 3, 5, 9, 17, 33 };
  return( nbpCode[iNbCode] );
}


int main(int argc, char *argv[])
{
  Param *param = new Param(); 
 
  printf("SYNOPSIS:\tcolorMapping\t-sx sizeX -sy sizeY -imaIn0 inputImage.yuv -imaIn1 inputImage.yuv -imaOut outputImage.yuv -inputBitDepth ibd -nbIma n -sei seiFileName \n\n");
  printf("\tmode=0\tImaInput => 1DlutA => 3x3 matrix + offset\n");
  printf("\tmode=1\tImaInput => 1DlutA => 3x3 matrix + offset => 1DlutB\n");
  printf("\ttype=n\t0:use float\t1:use integer\n");
  printf("\t\t\t-lutNbp\tnbp nodes in the 3D LUT (possible values = { 0, 2, 3, 5, 9, 17, 33 })\n");
  printf("\n\n");

  //==============================================================
  //  Initialization
  //==============================================================
  Int mode            = atoi( param->readOption(argc,argv,"mode","0") );
  Int m_inputBitDepth = atoi( param->readOption(argc,argv,"inputBitDepth","8") );
  Int m_outputBitDepth = atoi( param->readOption(argc,argv,"outputBitDepth","8") );
  Int m_appendBitDepth = atoi( param->readOption(argc,argv,"appendBitDepth","8") );
  Int m_lutBitDepth   = atoi( param->readOption(argc,argv,"lutBitDepth","8") );
  Int m_subSampling   = atoi( param->readOption(argc,argv,"subSampling","1") );
  Int m_border        = atoi( param->readOption(argc,argv,"border","16") );
  Int nbIma           = atoi( param->readOption(argc,argv,"nbIma","1") );
  Int nbIter          = atoi( param->readOption(argc,argv,"nbIter","5") );
  Int nbDeltaIma      = atoi( param->readOption(argc,argv,"nbDeltaIma","1") );
  Int model           = atoi( param->readOption(argc,argv,"model","1") );
  Int type            = atoi( param->readOption(argc,argv,"type","1") );
  Bool bVerbose       = param->isOption(argc,argv,"verbose");
  Int denomShift      = m_lutBitDepth - 8;

  Double    d_denomShift  = denomShift ? ((Double)(1<<denomShift)) : (1.0) ;
  Double    d_denomOffset = d_denomShift / 2.0;

  //==============================================================
  //  Collecting Color samples
  //==============================================================
  Int nbpCode = atoi(param->readOption(argc,argv,"lutNbp"));
  Int nbp     = getNb( nbpCode );


  printf("START\n");

  TComColorMappingMainAlgo* colorMappingMain  = new TComColorMappingMainAlgo( nbpCode, m_inputBitDepth, m_appendBitDepth, m_lutBitDepth, m_subSampling, m_border, bVerbose );
  TCom1DlutGO*              colorMapping1DlutGO = new TCom1DlutGO( nbpCode, m_inputBitDepth, m_appendBitDepth, m_lutBitDepth, m_subSampling, m_border, bVerbose );

  TComPicYuv* pcBasePic[3]  = {NULL,NULL,NULL};
  FILE* ficIn[3]            = {NULL,NULL,NULL};
  Int sizeX[3]              = {0, 0, 0};
  Int sizeY[3]              = {0, 0, 0};
  sizeX[0]  = atoi(param->readOption(argc,argv,"sx0"));
  sizeX[1]  = atoi(param->readOption(argc,argv,"sx1"));
  sizeY[0]  = atoi(param->readOption(argc,argv,"sy0"));
  sizeY[1]  = atoi(param->readOption(argc,argv,"sy1"));

  Bool  b2x = (sizeX[0] == (sizeX[1]>>1)) && (sizeY[0] == (sizeY[1]>>1)) ;
  Int marginX = b2x ? 16 : 0 ;
  Int marginY = b2x ? 16 : 0 ;

  for ( Int i=0 ; i<2 ; i++ ) {
    pcBasePic[i]  = new TComPicYuv;
    pcBasePic[i]->create( sizeX[i], sizeY[i], marginX, marginY );
  }
  if ( b2x ) { // Down-sampled picture
    pcBasePic[2]  = new TComPicYuv;
    pcBasePic[2]->create( sizeX[0], sizeY[0], marginX, marginY );
  }
  ficIn[0]  = fopen( param->readOption(argc,argv, "imaIn0"), "rb" );
  ficIn[1]  = fopen( param->readOption(argc,argv, "imaIn1"), "rb" );
  pcBasePic[0]->read( ficIn[0], 0, m_inputBitDepth );
  pcBasePic[1]->read( ficIn[1], 0, m_outputBitDepth );

  if ( b2x ) {
    printf("Subsampling...\n");
    colorMappingMain->downScalePic( pcBasePic[1], pcBasePic[2], m_outputBitDepth, m_inputBitDepth );

    Char  tmpFileNameOut[1000];
    buildNameSequence( "downScaled", NULL, sizeX[0], sizeY[0], 0, tmpFileNameOut, m_inputBitDepth );

    FILE* ficOut = fopen( tmpFileNameOut, "wb");
    pcBasePic[2]->write( ficOut, 0, m_inputBitDepth, m_inputBitDepth );

    fclose(ficOut);
  }

  colorMapping1DlutGO->compute1DLutGO( pcBasePic[0], pcBasePic[ b2x ? 2 : 1 ], model==1, nbIter );
  if ( param->isOption(argc,argv,"sei") ) {
    colorMapping1DlutGO->createSei( param->readOption(argc,argv,"sei"), model==1 );
  }

  TComPicYuv* pcOutPic  = new TComPicYuv;
  pcOutPic->create( sizeX[0], sizeY[0], marginX, marginY );
  colorMapping1DlutGO->colorMap( pcBasePic[0], pcOutPic, 10, model==1, type==1 );

  printf("Output sequence:\t%s\n", param->readOption(argc,argv, "imaOut") );
  FILE* ficOut = fopen( param->readOption(argc,argv, "imaOut"), "wb");
  //pcOutPic->write( ficOut, 0, m_inputBitDepth, m_inputBitDepth );
  pcOutPic->write( ficOut, 0, 10, 10 );
  fclose(ficOut);
}
